# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
The L{_response} module contains constants for all standard HTTP codes, along
with a mapping to the corresponding phrases.
"""


import string

from twisted.trial import unittest
from twisted.web import _responses


class ResponseTests(unittest.TestCase):
    def test_constants(self) -> None:
        """
        All constants besides C{RESPONSES} defined in L{_response} are
        integers and are keys in C{RESPONSES}.
        """
        for sym in dir(_responses):
            if sym == "RESPONSES":
                continue
            if all((c == "_" or c in string.ascii_uppercase) for c in sym):
                val = getattr(_responses, sym)
                self.assertIsInstance(val, int)
                self.assertIn(val, _responses.RESPONSES)



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Tests for large portions of L{twisted.mail}.
"""
from __future__ import annotations

import email.message
import email.parser
import errno
import glob
import io
import os
import pickle
import shutil
import signal
import sys
import tempfile
import textwrap
import time
from hashlib import md5
from typing import Any, NoReturn
from unittest import skipIf

from zope.interface import Interface, implementer
from zope.interface.verify import verifyClass

import twisted.cred.checkers
import twisted.cred.credentials
import twisted.cred.portal
import twisted.mail.alias
import twisted.mail.mail
import twisted.mail.maildir
import twisted.mail.protocols
import twisted.mail.relay
import twisted.mail.relaymanager
from twisted import cred, mail
from twisted.internet import address, defer, interfaces, protocol, reactor, task
from twisted.internet.defer import Deferred
from twisted.internet.error import (
    CannotListenError,
    DNSLookupError,
    ProcessDone,
    ProcessTerminated,
)
from twisted.internet.testing import (
    LineSendingProtocol,
    MemoryReactorClock,
    StringTransport,
)
from twisted.mail import pop3, smtp
from twisted.mail.maildir import MaildirMailbox
from twisted.mail.relaymanager import _AttemptManager
from twisted.names import dns
from twisted.names.dns import Record_CNAME, Record_MX, RRHeader
from twisted.names.error import DNSNameError
from twisted.python import failure, log
from twisted.python.filepath import FilePath
from twisted.python.runtime import platformType
from twisted.trial.unittest import TestCase


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class DomainWithDefaultsTests(TestCase):
    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testMethods(self):
        d = {x: x + 10 for x in range(10)}
        d = mail.mail.DomainWithDefaultDict(d, "Default")

        self.assertEqual(len(d), 10)
        self.assertEqual(list(iter(d)), list(range(10)))
        self.assertEqual(list(d.iterkeys()), list(iter(d)))

        items = list(d.iteritems())
        items.sort()
        self.assertEqual(items, [(x, x + 10) for x in range(10)])

        values = list(d.itervalues())
        values.sort()
        self.assertEqual(values, list(range(10, 20)))

        items = d.items()
        items.sort()
        self.assertEqual(items, [(x, x + 10) for x in range(10)])

        values = d.values()
        values.sort()
        self.assertEqual(values, list(range(10, 20)))

        for x in range(10):
            self.assertEqual(d[x], x + 10)
            self.assertEqual(d.get(x), x + 10)
            self.assertTrue(x in d)

        del d[2], d[4], d[6]

        self.assertEqual(len(d), 7)
        self.assertEqual(d[2], "Default")
        self.assertEqual(d[4], "Default")
        self.assertEqual(d[6], "Default")

        d.update({"a": None, "b": (), "c": "*"})
        self.assertEqual(len(d), 10)
        self.assertEqual(d["a"], None)
        self.assertEqual(d["b"], ())
        self.assertEqual(d["c"], "*")

        d.clear()
        self.assertEqual(len(d), 0)

        self.assertEqual(d.setdefault("key", "value"), "value")
        self.assertEqual(d["key"], "value")

        self.assertEqual(d.popitem(), ("key", "value"))
        self.assertEqual(len(d), 0)

        dcopy = d.copy()
        self.assertEqual(d.domains, dcopy.domains)
        self.assertEqual(d.default, dcopy.default)

    def _stringificationTest(self, stringifier):
        """
        Assert that the class name of a L{mail.mail.DomainWithDefaultDict}
        instance and the string-formatted underlying domain dictionary both
        appear in the string produced by the given string-returning function.

        @type stringifier: one-argument callable
        @param stringifier: either C{str} or C{repr}, to be used to get a
            string to make assertions against.
        """
        domain = mail.mail.DomainWithDefaultDict({}, "Default")
        self.assertIn(domain.__class__.__name__, stringifier(domain))
        domain["key"] = "value"
        self.assertIn(str({"key": "value"}), stringifier(domain))

    def test_str(self):
        """
        L{DomainWithDefaultDict.__str__} should return a string including
        the class name and the domain mapping held by the instance.
        """
        self._stringificationTest(str)

    def test_repr(self):
        """
        L{DomainWithDefaultDict.__repr__} should return a string including
        the class name and the domain mapping held by the instance.
        """
        self._stringificationTest(repr)

    def test_has_keyDeprecation(self):
        """
        has_key is now deprecated.
        """
        sut = mail.mail.DomainWithDefaultDict({}, "Default")

        sut.has_key("anything")

        message = (
            "twisted.mail.mail.DomainWithDefaultDict.has_key was deprecated "
            "in Twisted 16.3.0. Use the `in` keyword instead."
        )
        warnings = self.flushWarnings([self.test_has_keyDeprecation])
        self.assertEqual(1, len(warnings))
        self.assertEqual(DeprecationWarning, warnings[0]["category"])
        self.assertEqual(message, warnings[0]["message"])


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class BounceTests(TestCase):
    def setUp(self):
        self.domain = mail.mail.BounceDomain()

    def testExists(self):
        self.assertRaises(smtp.AddressError, self.domain.exists, "any user")

    def testRelay(self):
        self.assertEqual(self.domain.willRelay("random q emailer", "protocol"), False)

    def testAddUser(self):
        self.domain.addUser("bob", "password")
        self.assertRaises(smtp.SMTPBadRcpt, self.domain.exists, "bob")


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class BounceWithSMTPServerTests(TestCase):
    """
    Tests for L{twisted.mail.mail.BounceDomain} with
    L{twisted.mail.smtp.SMTPServer}.
    """

    def test_rejected(self):
        """
        Incoming emails to a SMTP server with L{twisted.mail.mail.BounceDomain}
        are rejected.
        """
        service = mail.mail.MailService()
        domain = mail.mail.BounceDomain()
        service.addDomain(b"foo.com", domain)

        factory = mail.protocols.SMTPFactory(service)
        protocol = factory.buildProtocol(None)

        deliverer = mail.protocols.SMTPDomainDelivery(service, None, None)
        protocol.delivery = deliverer

        transport = StringTransport()
        protocol.makeConnection(transport)

        protocol.lineReceived(b"HELO baz.net")
        protocol.lineReceived(b"MAIL FROM:<a@baz.net>")
        protocol.lineReceived(b"RCPT TO:<any@foo.com>")
        protocol.lineReceived(b"QUIT")

        self.assertTrue(transport.disconnecting)
        protocol.connectionLost(None)

        self.assertEqual(
            transport.value().strip().split(b"\r\n")[-2],
            b"550 Cannot receive for specified address",
        )


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class FileMessageTests(TestCase):
    def setUp(self):
        self.name = self.mktemp()
        self.final = self.mktemp()
        self.f = open(self.name, "wb")
        self.addCleanup(self.f.close)
        self.fp = mail.mail.FileMessage(self.f, self.name, self.final)

    def testFinalName(self):
        return self.fp.eomReceived().addCallback(self._cbFinalName)

    def _cbFinalName(self, result):
        self.assertEqual(result, self.final)
        self.assertTrue(self.f.closed)
        self.assertFalse(os.path.exists(self.name))

    def testContents(self):
        contents = b"first line\nsecond line\nthird line\n"
        for line in contents.splitlines():
            self.fp.lineReceived(line)
        self.fp.eomReceived()
        with open(self.final, "rb") as f:
            self.assertEqual(f.read(), contents)

    def testInterrupted(self):
        contents = b"first line\nsecond line\n"
        for line in contents.splitlines():
            self.fp.lineReceived(line)
        self.fp.connectionLost()
        self.assertFalse(os.path.exists(self.name))
        self.assertFalse(os.path.exists(self.final))


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class MaildirMessageTests(TestCase):
    """
    Tests for the file creating by the L{mail.maildir.MaildirMessage}.
    """

    def setUp(self):
        """
        Create and open a temporary file.
        """
        self.name = self.mktemp()
        self.final = self.mktemp()
        self.address = b"user@example.com"
        self.f = open(self.name, "wb")
        self.addCleanup(self.f.close)
        self.fp = mail.maildir.MaildirMessage(
            self.address, self.f, self.name, self.final
        )

    def _finalName(self):
        """
        Search for the final file path.

        @rtype: L{str}
        @return: Final file path.
        """
        return glob.glob(f"{self.final},S=[0-9]*")[0]

    def test_finalName(self):
        """
        Send the EOM to the message and check that the final file name contains
        the correct file size and the temporary file has been closed and removed.
        """
        final = self.successResultOf(self.fp.eomReceived())
        self.assertEqual(final, f"{self.final},S={os.path.getsize(final)}")
        self.assertTrue(self.f.closed)
        self.assertFalse(os.path.exists(self.name))

    def test_contents(self):
        """
        Send a message contents and the EOM to the message and check that the
        final file contains the correct header and the message contents.
        """
        contents = b"first line\nsecond line\nthird line\n"
        for line in contents.splitlines():
            self.fp.lineReceived(line)
        final = self.successResultOf(self.fp.eomReceived())
        with open(final, "rb") as f:
            self.assertEqual(
                f.read(), b"Delivered-To: %s\n%s" % (self.address, contents)
            )

    def test_interrupted(self):
        """
        Check that the interrupted message transfer removes the temporary file
        and a doesn't create a final file.
        """
        contents = b"first line\nsecond line\n"
        for line in contents.splitlines():
            self.fp.lineReceived(line)
        self.fp.connectionLost()
        self.assertFalse(os.path.exists(self.name))
        self.assertRaises(IndexError, self._finalName)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class MailServiceTests(TestCase):
    def setUp(self):
        self.service = mail.mail.MailService()

    def testFactories(self):
        f = self.service.getPOP3Factory()
        self.assertTrue(isinstance(f, protocol.ServerFactory))
        self.assertTrue(f.buildProtocol(("127.0.0.1", 12345)), pop3.POP3)

        f = self.service.getSMTPFactory()
        self.assertTrue(isinstance(f, protocol.ServerFactory))
        self.assertTrue(f.buildProtocol(("127.0.0.1", 12345)), smtp.SMTP)

        f = self.service.getESMTPFactory()
        self.assertTrue(isinstance(f, protocol.ServerFactory))
        self.assertTrue(f.buildProtocol(("127.0.0.1", 12345)), smtp.ESMTP)

    def testPortals(self):
        o1 = object()
        o2 = object()
        self.service.portals["domain"] = o1
        self.service.portals[""] = o2

        self.assertTrue(self.service.lookupPortal("domain") is o1)
        self.assertTrue(self.service.defaultPortal() is o2)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class StringListMailboxTests(TestCase):
    """
    Tests for L{StringListMailbox}, an in-memory only implementation of
    L{pop3.IMailbox}.
    """

    def test_listOneMessage(self):
        """
        L{StringListMailbox.listMessages} returns the length of the message at
        the offset into the mailbox passed to it.
        """
        mailbox = mail.maildir.StringListMailbox(["abc", "ab", "a"])
        self.assertEqual(mailbox.listMessages(0), 3)
        self.assertEqual(mailbox.listMessages(1), 2)
        self.assertEqual(mailbox.listMessages(2), 1)

    def test_listAllMessages(self):
        """
        L{StringListMailbox.listMessages} returns a list of the lengths of all
        messages if not passed an index.
        """
        mailbox = mail.maildir.StringListMailbox(["a", "abc", "ab"])
        self.assertEqual(mailbox.listMessages(), [1, 3, 2])

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_getMessage(self):
        """
        L{StringListMailbox.getMessage} returns a file-like object from which
        the contents of the message at the given offset into the mailbox can be
        read.
        """
        mailbox = mail.maildir.StringListMailbox(["foo", "real contents"])
        self.assertEqual(mailbox.getMessage(1).read(), "real contents")

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_getUidl(self):
        """
        L{StringListMailbox.getUidl} returns a unique identifier for the
        message at the given offset into the mailbox.
        """
        mailbox = mail.maildir.StringListMailbox(["foo", "bar"])
        self.assertNotEqual(mailbox.getUidl(0), mailbox.getUidl(1))

    def test_deleteMessage(self):
        """
        L{StringListMailbox.deleteMessage} marks a message for deletion causing
        further requests for its length to return 0.
        """
        mailbox = mail.maildir.StringListMailbox(["foo"])
        mailbox.deleteMessage(0)
        self.assertEqual(mailbox.listMessages(0), 0)
        self.assertEqual(mailbox.listMessages(), [0])

    def test_undeleteMessages(self):
        """
        L{StringListMailbox.undeleteMessages} causes any messages marked for
        deletion to be returned to their original state.
        """
        mailbox = mail.maildir.StringListMailbox(["foo"])
        mailbox.deleteMessage(0)
        mailbox.undeleteMessages()
        self.assertEqual(mailbox.listMessages(0), 3)
        self.assertEqual(mailbox.listMessages(), [3])

    def test_sync(self):
        """
        L{StringListMailbox.sync} causes any messages as marked for deletion to
        be permanently deleted.
        """
        mailbox = mail.maildir.StringListMailbox(["foo"])
        mailbox.deleteMessage(0)
        mailbox.sync()
        mailbox.undeleteMessages()
        self.assertEqual(mailbox.listMessages(0), 0)
        self.assertEqual(mailbox.listMessages(), [0])


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class FailingMaildirMailboxAppendMessageTask(
    mail.maildir._MaildirMailboxAppendMessageTask
):
    _openstate = True
    _writestate = True
    _renamestate = True

    def osopen(self, fn, attr, mode):
        if self._openstate:
            return os.open(fn, attr, mode)
        else:
            raise OSError(errno.EPERM, "Faked Permission Problem")

    def oswrite(self, fh, data):
        if self._writestate:
            return os.write(fh, data)
        else:
            raise OSError(errno.ENOSPC, "Faked Space problem")

    def osrename(self, oldname, newname):
        if self._renamestate:
            return os.rename(oldname, newname)
        else:
            raise OSError(errno.EPERM, "Faked Permission Problem")


class _AppendTestMixin:
    """
    Mixin for L{MaildirMailbox.appendMessage} test cases which defines a helper
    for serially appending multiple messages to a mailbox.
    """

    def _appendMessages(
        self, mbox: twisted.mail.maildir.MaildirMailbox, messages: list[bytes]
    ) -> Deferred[list[None]]:
        """
        Deliver the given messages one at a time.  Delivery is serialized to
        guarantee a predictable order in the mailbox (overlapped message deliver
        makes no guarantees about which message which appear first).
        """

        async def append() -> list[None]:
            results: list[None] = []
            for m in messages:
                await mbox.appendMessage(m)
                results.append(None)
            return results

        return Deferred.fromCoroutine(append())


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class MaildirAppendStringTests(TestCase, _AppendTestMixin):
    """
    Tests for L{MaildirMailbox.appendMessage} when invoked with a C{str}.
    """

    def setUp(self) -> None:
        self.d = os.fsencode(self.mktemp())
        mail.maildir.initializeMaildir(self.d)

    def _append(self, ignored: object, mbox: MaildirMailbox) -> Deferred[None]:
        d = mbox.appendMessage(b"TEST")
        return self.assertFailure(d, Exception)

    def _setState(self, ignored, mbox, rename=None, write=None, open=None):
        """
        Change the behavior of future C{rename}, C{write}, or C{open} calls made
        by the mailbox C{mbox}.

        @param rename: If not L{None}, a new value for the C{_renamestate}
            attribute of the mailbox's append factory.  The original value will
            be restored at the end of the test.

        @param write: Like C{rename}, but for the C{_writestate} attribute.

        @param open: Like C{rename}, but for the C{_openstate} attribute.
        """
        if rename is not None:
            self.addCleanup(
                setattr,
                mbox.AppendFactory,
                "_renamestate",
                mbox.AppendFactory._renamestate,
            )
            mbox.AppendFactory._renamestate = rename
        if write is not None:
            self.addCleanup(
                setattr,
                mbox.AppendFactory,
                "_writestate",
                mbox.AppendFactory._writestate,
            )
            mbox.AppendFactory._writestate = write
        if open is not None:
            self.addCleanup(
                setattr, mbox.AppendFactory, "_openstate", mbox.AppendFactory._openstate
            )
            mbox.AppendFactory._openstate = open

    def test_append(self) -> Any:
        """
        L{MaildirMailbox.appendMessage} returns a L{Deferred} which fires when
        the message has been added to the end of the mailbox.
        """
        mbox = mail.maildir.MaildirMailbox(self.d)
        mbox.AppendFactory = FailingMaildirMailboxAppendMessageTask

        d = self._appendMessages(mbox, [b"X" * i for i in range(1, 11)])
        d.addCallback(self.assertEqual, [None] * 10)
        d.addCallback(self._cbTestAppend, mbox)
        return d

    def test_reopenMailbox(self) -> Any:
        """
        Appending a message, then re-opening the mailbox by creating a new
        MaildirMailbox, should result in seeing the message as loadable.
        """

        async def go() -> None:
            mbox1 = mail.maildir.MaildirMailbox(self.d)
            sample = b"xyzzy"
            await mbox1.appendMessage(sample)
            mbox2 = mail.maildir.MaildirMailbox(self.d)
            listed = mbox2.listMessages()
            self.assertEqual(len(listed), 1)
            [messageSize] = listed
            self.assertEqual(messageSize, len(sample))
            self.assertEqual(mbox2.getMessage(0).read(), sample.decode())

        return Deferred.fromCoroutine(go())

    def _cbTestAppend(self, ignored, mbox):
        """
        Check that the mailbox has the expected number (ten) of messages in it,
        and that each has the expected contents, and that they are in the same
        order as that in which they were appended.
        """
        self.assertEqual(len(mbox.listMessages()), 10)
        self.assertEqual(
            [len(mbox.getMessage(i).read()) for i in range(10)], list(range(1, 11))
        )
        # test in the right order: last to first error location.
        self._setState(None, mbox, rename=False)
        d = self._append(None, mbox)
        d.addCallback(self._setState, mbox, rename=True, write=False)
        d.addCallback(self._append, mbox)
        d.addCallback(self._setState, mbox, write=True, open=False)
        d.addCallback(self._append, mbox)
        d.addCallback(self._setState, mbox, open=True)
        return d


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class MaildirAppendFileTests(TestCase, _AppendTestMixin):
    """
    Tests for L{MaildirMailbox.appendMessage} when invoked with a C{str}.
    """

    def setUp(self):
        self.d = self.mktemp()
        mail.maildir.initializeMaildir(self.d)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_append(self):
        """
        L{MaildirMailbox.appendMessage} returns a L{Deferred} which fires when
        the message has been added to the end of the mailbox.
        """
        mbox = mail.maildir.MaildirMailbox(self.d)
        messages = []
        for i in range(1, 11):
            temp = tempfile.TemporaryFile()
            temp.write("X" * i)
            temp.seek(0, 0)
            messages.append(temp)
            self.addCleanup(temp.close)

        d = self._appendMessages(mbox, messages)
        d.addCallback(self._cbTestAppend, mbox)
        return d

    def _cbTestAppend(self, result, mbox):
        """
        Check that the mailbox has the expected number (ten) of messages in it,
        and that each has the expected contents, and that they are in the same
        order as that in which they were appended.
        """
        self.assertEqual(len(mbox.listMessages()), 10)
        self.assertEqual(
            [len(mbox.getMessage(i).read()) for i in range(10)], list(range(1, 11))
        )


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class MaildirTests(TestCase):
    def setUp(self):
        self.d = self.mktemp()
        mail.maildir.initializeMaildir(self.d)

    def tearDown(self):
        shutil.rmtree(self.d)

    def testInitializer(self):
        d = self.d
        trash = os.path.join(d, ".Trash")

        self.assertTrue(os.path.exists(d) and os.path.isdir(d))
        self.assertTrue(os.path.exists(os.path.join(d, "new")))
        self.assertTrue(os.path.exists(os.path.join(d, "cur")))
        self.assertTrue(os.path.exists(os.path.join(d, "tmp")))
        self.assertTrue(os.path.isdir(os.path.join(d, "new")))
        self.assertTrue(os.path.isdir(os.path.join(d, "cur")))
        self.assertTrue(os.path.isdir(os.path.join(d, "tmp")))

        self.assertTrue(os.path.exists(os.path.join(trash, "new")))
        self.assertTrue(os.path.exists(os.path.join(trash, "cur")))
        self.assertTrue(os.path.exists(os.path.join(trash, "tmp")))
        self.assertTrue(os.path.isdir(os.path.join(trash, "new")))
        self.assertTrue(os.path.isdir(os.path.join(trash, "cur")))
        self.assertTrue(os.path.isdir(os.path.join(trash, "tmp")))

    def test_nameGenerator(self):
        """
        Each call to L{_MaildirNameGenerator.generate} returns a unique
        string suitable for use as the basename of a new message file.  The
        names are ordered such that those generated earlier sort less than
        those generated later.
        """
        clock = task.Clock()
        clock.advance(0.05)
        generator = mail.maildir._MaildirNameGenerator(clock)

        firstName = generator.generate()
        clock.advance(0.05)
        secondName = generator.generate()

        self.assertTrue(firstName < secondName)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_mailbox(self):
        """
        Exercise the methods of L{IMailbox} as implemented by
        L{MaildirMailbox}.
        """
        j = os.path.join
        n = mail.maildir._generateMaildirName
        msgs = [j(b, n()) for b in ("cur", "new") for x in range(5)]

        # Toss a few files into the mailbox
        i = 1
        for f in msgs:
            with open(j(self.d, f), "w") as fObj:
                fObj.write("x" * i)
            i = i + 1

        mb = mail.maildir.MaildirMailbox(self.d)
        self.assertEqual(mb.listMessages(), list(range(1, 11)))
        self.assertEqual(mb.listMessages(1), 2)
        self.assertEqual(mb.listMessages(5), 6)

        self.assertEqual(mb.getMessage(6).read(), "x" * 7)
        self.assertEqual(mb.getMessage(1).read(), "x" * 2)

        d = {}
        for i in range(10):
            u = mb.getUidl(i)
            self.assertFalse(u in d)
            d[u] = None

        p, f = os.path.split(msgs[5])

        mb.deleteMessage(5)
        self.assertEqual(mb.listMessages(5), 0)
        self.assertTrue(os.path.exists(j(self.d, ".Trash", "cur", f)))
        self.assertFalse(os.path.exists(j(self.d, msgs[5])))

        mb.undeleteMessages()
        self.assertEqual(mb.listMessages(5), 6)
        self.assertFalse(os.path.exists(j(self.d, ".Trash", "cur", f)))
        self.assertTrue(os.path.exists(j(self.d, msgs[5])))


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class AbstractMaildirDomainTests(TestCase):
    """
    Tests for L{twisted.mail.maildir.AbstractMaildirDomain}.
    """

    def test_interface(self):
        """
        L{maildir.AbstractMaildirDomain} implements L{mail.IAliasableDomain}.
        """
        verifyClass(mail.mail.IAliasableDomain, mail.maildir.AbstractMaildirDomain)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class MaildirDirdbmDomainTests(TestCase):
    """
    Tests for L{MaildirDirdbmDomain}.
    """

    def setUp(self) -> None:
        """
        Create a temporary L{MaildirDirdbmDomain} and parent
        L{MailService} before running each test.
        """
        self.P = self.mktemp()
        self.S = mail.mail.MailService()
        self.D = mail.maildir.MaildirDirdbmDomain(self.S, self.P)

    def tearDown(self):
        """
        Remove the temporary C{maildir} directory when the test has
        finished.
        """
        shutil.rmtree(self.P)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_addUser(self):
        """
        L{MaildirDirdbmDomain.addUser} accepts a user and password
        argument. It stores those in a C{dbm} dictionary
        attribute and creates a directory for each user.
        """
        toAdd = (("user1", "pwd1"), ("user2", "pwd2"), ("user3", "pwd3"))
        for u, p in toAdd:
            self.D.addUser(u, p)

        for u, p in toAdd:
            self.assertTrue(u in self.D.dbm)
            self.assertEqual(self.D.dbm[u], p)
            self.assertTrue(os.path.exists(os.path.join(self.P, u)))

    def test_credentials(self):
        """
        L{MaildirDirdbmDomain.getCredentialsCheckers} initializes and
        returns one L{ICredentialsChecker} checker by default.
        """
        creds = self.D.getCredentialsCheckers()

        self.assertEqual(len(creds), 1)
        self.assertTrue(cred.checkers.ICredentialsChecker.providedBy(creds[0]))
        self.assertTrue(
            cred.credentials.IUsernamePassword in creds[0].credentialInterfaces
        )

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_requestAvatar(self):
        """
        L{MaildirDirdbmDomain.requestAvatar} raises L{NotImplementedError}
        unless it is supplied with an L{pop3.IMailbox} interface.
        When called with an L{pop3.IMailbox}, it returns a 3-tuple
        containing L{pop3.IMailbox}, an implementation of that interface
        and a NOOP callable.
        """

        class ISomething(Interface):
            pass

        self.D.addUser("user", "password")
        self.assertRaises(
            NotImplementedError, self.D.requestAvatar, "user", None, ISomething
        )

        t = self.D.requestAvatar("user", None, pop3.IMailbox)
        self.assertEqual(len(t), 3)
        self.assertTrue(t[0] is pop3.IMailbox)
        self.assertTrue(pop3.IMailbox.providedBy(t[1]))

        t[2]()

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_requestAvatarId(self):
        """
        L{DirdbmDatabase.requestAvatarId} raises L{UnauthorizedLogin} if
        supplied with invalid user credentials.
        When called with valid credentials, L{requestAvatarId} returns
        the username associated with the supplied credentials.
        """
        self.D.addUser("user", "password")
        database = self.D.getCredentialsCheckers()[0]

        creds = cred.credentials.UsernamePassword("user", "wrong password")
        self.assertRaises(cred.error.UnauthorizedLogin, database.requestAvatarId, creds)

        creds = cred.credentials.UsernamePassword("user", "password")
        self.assertEqual(database.requestAvatarId(creds), "user")

    def test_userDirectory(self) -> None:
        """
        L{MaildirDirdbmDomain.userDirectory} is supplied with a user name
        and returns the path to that user's maildir subdirectory.
        Calling L{MaildirDirdbmDomain.userDirectory} with a
        non-existent user returns the 'postmaster' directory if there
        is a postmaster or returns L{None} if there is no postmaster.
        """
        self.D.addUser(b"user", b"password")
        self.assertEqual(
            self.D.userDirectory(b"user"), os.path.join(self.D.root, b"user")
        )

        self.D.postmaster = False
        self.assertIdentical(self.D.userDirectory(b"nouser"), None)

        self.D.postmaster = True
        self.assertEqual(
            self.D.userDirectory(b"nouser"),
            os.path.join(self.D.root, b"postmaster"),
        )


@implementer(mail.mail.IAliasableDomain)
class StubAliasableDomain:
    """
    Minimal testable implementation of IAliasableDomain.
    """

    def exists(self, user, memo=None):
        """
        No test coverage for invocations of this method on domain objects,
        so we just won't implement it.
        """
        raise NotImplementedError()

    def addUser(self, user, password):
        """
        No test coverage for invocations of this method on domain objects,
        so we just won't implement it.
        """
        raise NotImplementedError()

    def getCredentialsCheckers(self):
        """
        This needs to succeed in order for other tests to complete
        successfully, but we don't actually assert anything about its
        behavior.  Return an empty list.  Sometime later we should return
        something else and assert that a portal got set up properly.
        """
        return []

    def setAliasGroup(self, aliases):
        """
        Just record the value so the test can check it later.
        """
        self.aliasGroup = aliases

    def requestAvatar(
        self, avatarId: bytes | tuple[()], mind: object, *interfaces: type[Interface]
    ) -> NoReturn:
        """
        Stub domains cannot authenticate users.
        """
        raise NotImplementedError()


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class ServiceDomainTests(TestCase):
    def setUp(self):
        self.S = mail.mail.MailService()
        self.D = mail.protocols.DomainDeliveryBase(self.S, None)
        self.D.service = self.S
        self.D.protocolName = "TEST"
        self.D.host = "hostname"

        self.tmpdir = self.mktemp()
        domain = mail.maildir.MaildirDirdbmDomain(self.S, self.tmpdir)
        domain.addUser(b"user", b"password")
        self.S.addDomain("test.domain", domain)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def testAddAliasableDomain(self):
        """
        Test that adding an IAliasableDomain to a mail service properly sets
        up alias group references and such.
        """
        aliases = object()
        domain = StubAliasableDomain()
        self.S.aliases = aliases
        self.S.addDomain("example.com", domain)
        self.assertIdentical(domain.aliasGroup, aliases)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testReceivedHeader(self):
        hdr = self.D.receivedHeader(
            ("remotehost", "123.232.101.234"),
            smtp.Address("<someguy@someplace>"),
            ["user@host.name"],
        )
        fp = io.BytesIO(hdr)
        emailParser = email.parser.Parser()
        m = emailParser.parse(fp)
        self.assertEqual(len(m.items()), 1)
        self.assertIn("Received", m)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testValidateTo(self):
        user = smtp.User("user@test.domain", "helo", None, "wherever@whatever")
        return defer.maybeDeferred(self.D.validateTo, user).addCallback(
            self._cbValidateTo
        )

    def _cbValidateTo(self, result):
        self.assertTrue(callable(result))

    def testValidateToBadUsername(self):
        user = smtp.User("resu@test.domain", "helo", None, "wherever@whatever")
        return self.assertFailure(
            defer.maybeDeferred(self.D.validateTo, user), smtp.SMTPBadRcpt
        )

    def testValidateToBadDomain(self):
        user = smtp.User("user@domain.test", "helo", None, "wherever@whatever")
        return self.assertFailure(
            defer.maybeDeferred(self.D.validateTo, user), smtp.SMTPBadRcpt
        )

    def testValidateFrom(self):
        helo = ("hostname", "127.0.0.1")
        origin = smtp.Address("<user@hostname>")
        self.assertTrue(self.D.validateFrom(helo, origin) is origin)

        helo = ("hostname", "1.2.3.4")
        origin = smtp.Address("<user@hostname>")
        self.assertTrue(self.D.validateFrom(helo, origin) is origin)

        helo = ("hostname", "1.2.3.4")
        origin = smtp.Address("<>")
        self.assertTrue(self.D.validateFrom(helo, origin) is origin)

        self.assertRaises(smtp.SMTPBadSender, self.D.validateFrom, None, origin)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class VirtualPOP3Tests(TestCase):
    def setUp(self) -> None:
        self.tmpdir = self.mktemp()
        self.S = mail.mail.MailService()
        self.D = mail.maildir.MaildirDirdbmDomain(self.S, self.tmpdir)
        self.D.addUser(b"user", b"password")
        self.S.addDomain(b"test.domain", self.D)
        self.S.addDomain(b"", self.D)
        self.P = mail.protocols.VirtualPOP3()
        self.P.service = self.S
        self.P.magic = b"<unit test magic>"

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testAuthenticateAPOP(self):
        resp = md5(self.P.magic + "password").hexdigest()
        return self.P.authenticateUserAPOP("user", resp).addCallback(
            self._cbAuthenticateAPOP
        )

    def _cbAuthenticateAPOP(self, result):
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], pop3.IMailbox)
        self.assertTrue(pop3.IMailbox.providedBy(result[1]))
        result[2]()

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testAuthenticateIncorrectUserAPOP(self):
        resp = md5(self.P.magic + "password").hexdigest()
        return self.assertFailure(
            self.P.authenticateUserAPOP("resu", resp), cred.error.UnauthorizedLogin
        )

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testAuthenticateIncorrectResponseAPOP(self):
        resp = md5("wrong digest").hexdigest()
        return self.assertFailure(
            self.P.authenticateUserAPOP("user", resp), cred.error.UnauthorizedLogin
        )

    def testAuthenticatePASS(self) -> Deferred[object]:
        return self.P.authenticateUserPASS(b"user", b"password").addCallback(
            self._cbAuthenticatePASS
        )

    def _cbAuthenticatePASS(self, result):
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], pop3.IMailbox)
        self.assertTrue(pop3.IMailbox.providedBy(result[1]))
        result[2]()

    def testAuthenticateBadUserPASS(self) -> Any:
        return self.assertFailure(
            self.P.authenticateUserPASS(b"resu", b"password"),
            cred.error.UnauthorizedLogin,
        )

    def testAuthenticateBadPasswordPASS(self) -> Any:
        return self.assertFailure(
            self.P.authenticateUserPASS(b"user", b"wrong password"),
            cred.error.UnauthorizedLogin,
        )


class empty(smtp.User):
    def __init__(self):
        pass


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class RelayTests(TestCase):
    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testExists(self):
        service = mail.mail.MailService()
        domain = mail.relay.DomainQueuer(service)

        doRelay = [
            address.UNIXAddress("/var/run/mail-relay"),
            address.IPv4Address("TCP", "127.0.0.1", 12345),
        ]

        dontRelay = [
            address.IPv4Address("TCP", "192.168.2.1", 62),
            address.IPv4Address("TCP", "1.2.3.4", 1943),
        ]

        for peer in doRelay:
            user = empty()
            user.orig = "user@host"
            user.dest = "tsoh@resu"
            user.protocol = empty()
            user.protocol.transport = empty()
            user.protocol.transport.getPeer = lambda: peer

            self.assertTrue(callable(domain.exists(user)))

        for peer in dontRelay:
            user = empty()
            user.orig = "some@place"
            user.protocol = empty()
            user.protocol.transport = empty()
            user.protocol.transport.getPeer = lambda: peer
            user.dest = "who@cares"

            self.assertRaises(smtp.SMTPBadRcpt, domain.exists, user)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class RelayerTests(TestCase):
    def setUp(self):
        self.tmpdir = self.mktemp()
        os.mkdir(self.tmpdir)
        self.messageFiles = []
        for i in range(10):
            name = os.path.join(self.tmpdir, "body-%d" % (i,))
            with open(name + "-H", "wb") as f:
                pickle.dump(["from-%d" % (i,), "to-%d" % (i,)], f)

            f = open(name + "-D", "w")
            f.write(name)
            f.seek(0, 0)
            self.messageFiles.append(name)

        self.R = mail.relay.RelayerMixin()
        self.R.loadMessages(self.messageFiles)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def testMailFrom(self):
        for i in range(10):
            self.assertEqual(self.R.getMailFrom(), "from-%d" % (i,))
            self.R.sentMail(250, None, None, None, None)
        self.assertEqual(self.R.getMailFrom(), None)

    def testMailTo(self):
        for i in range(10):
            self.assertEqual(self.R.getMailTo(), ["to-%d" % (i,)])
            self.R.sentMail(250, None, None, None, None)
        self.assertEqual(self.R.getMailTo(), None)

    def testMailData(self):
        for i in range(10):
            name = os.path.join(self.tmpdir, "body-%d" % (i,))
            self.assertEqual(self.R.getMailData().read(), name)
            self.R.sentMail(250, None, None, None, None)
        self.assertEqual(self.R.getMailData(), None)


class Manager:
    def __init__(self):
        self.success = []
        self.failure = []
        self.done = []

    def notifySuccess(self, factory, message):
        self.success.append((factory, message))

    def notifyFailure(self, factory, message):
        self.failure.append((factory, message))

    def notifyDone(self, factory):
        self.done.append(factory)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class ManagedRelayerTests(TestCase):
    def setUp(self):
        self.manager = Manager()
        self.messages = list(range(0, 20, 2))
        self.factory = object()
        self.relay = mail.relaymanager.ManagedRelayerMixin(self.manager)
        self.relay.messages = self.messages[:]
        self.relay.names = self.messages[:]
        self.relay.factory = self.factory

    def testSuccessfulSentMail(self):
        for i in self.messages:
            self.relay.sentMail(250, None, None, None, None)

        self.assertEqual(
            self.manager.success, [(self.factory, m) for m in self.messages]
        )

    def testFailedSentMail(self):
        for i in self.messages:
            self.relay.sentMail(550, None, None, None, None)

        self.assertEqual(
            self.manager.failure, [(self.factory, m) for m in self.messages]
        )

    def testConnectionLost(self):
        self.relay.connectionLost(failure.Failure(Exception()))
        self.assertEqual(self.manager.done, [self.factory])


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class DirectoryQueueTests(TestCase):
    def setUp(self):
        # This is almost a test case itself.
        self.tmpdir = self.mktemp()
        os.mkdir(self.tmpdir)
        self.queue = mail.relaymanager.Queue(self.tmpdir)
        self.queue.noisy = False
        for m in range(25):
            hdrF, msgF = self.queue.createNewMessage()
            with hdrF:
                pickle.dump(["header", m], hdrF)
            msgF.lineReceived(b"body: %d" % (m,))
            msgF.eomReceived()
        self.queue.readDirectory()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testWaiting(self):
        self.assertTrue(self.queue.hasWaiting())
        self.assertEqual(len(self.queue.getWaiting()), 25)

        waiting = self.queue.getWaiting()
        self.queue.setRelaying(waiting[0])
        self.assertEqual(len(self.queue.getWaiting()), 24)

        self.queue.setWaiting(waiting[0])
        self.assertEqual(len(self.queue.getWaiting()), 25)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testRelaying(self):
        for m in self.queue.getWaiting():
            self.queue.setRelaying(m)
            self.assertEqual(
                len(self.queue.getRelayed()), 25 - len(self.queue.getWaiting())
            )

        self.assertFalse(self.queue.hasWaiting())

        relayed = self.queue.getRelayed()
        self.queue.setWaiting(relayed[0])
        self.assertEqual(len(self.queue.getWaiting()), 1)
        self.assertEqual(len(self.queue.getRelayed()), 24)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testDone(self):
        msg = self.queue.getWaiting()[0]
        self.queue.setRelaying(msg)
        self.queue.done(msg)

        self.assertEqual(len(self.queue.getWaiting()), 24)
        self.assertEqual(len(self.queue.getRelayed()), 0)

        self.assertFalse(msg in self.queue.getWaiting())
        self.assertFalse(msg in self.queue.getRelayed())

    def testEnvelope(self):
        envelopes = []

        for msg in self.queue.getWaiting():
            envelopes.append(self.queue.getEnvelope(msg))

        envelopes.sort()
        for i in range(25):
            self.assertEqual(envelopes.pop(0), ["header", i])


from twisted.names import client, common, server


class TestAuthority(common.ResolverBase):
    def __init__(self):
        common.ResolverBase.__init__(self)
        self.addresses = {}

    def _lookup(self, name, cls, type, timeout=None):
        if name in self.addresses and type == dns.MX:
            results = []
            for a in self.addresses[name]:
                hdr = dns.RRHeader(name, dns.MX, dns.IN, 60, dns.Record_MX(0, a))
                results.append(hdr)
            return defer.succeed((results, [], []))
        return defer.fail(failure.Failure(dns.DomainError(name)))


def setUpDNS(self):
    self.auth = TestAuthority()
    factory = server.DNSServerFactory([self.auth])
    protocol = dns.DNSDatagramProtocol(factory)
    while 1:
        self.port = reactor.listenTCP(0, factory, interface="127.0.0.1")
        portNumber = self.port.getHost().port

        try:
            self.udpPort = reactor.listenUDP(
                portNumber, protocol, interface="127.0.0.1"
            )
        except CannotListenError:
            self.port.stopListening()
        else:
            break
    self.resolver = client.Resolver(servers=[("127.0.0.1", portNumber)])


def tearDownDNS(self):
    dl = []
    dl.append(defer.maybeDeferred(self.port.stopListening))
    dl.append(defer.maybeDeferred(self.udpPort.stopListening))
    try:
        self.resolver._parseCall.cancel()
    except BaseException:
        pass
    return defer.DeferredList(dl)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class MXTests(TestCase):
    """
    Tests for L{mail.relaymanager.MXCalculator}.
    """

    def setUp(self):
        setUpDNS(self)
        self.clock = task.Clock()
        self.mx = mail.relaymanager.MXCalculator(self.resolver, self.clock)

    def tearDown(self):
        return tearDownDNS(self)

    def test_defaultClock(self):
        """
        L{MXCalculator}'s default clock is C{twisted.internet.reactor}.
        """
        self.assertIdentical(
            mail.relaymanager.MXCalculator(self.resolver).clock, reactor
        )

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testSimpleSuccess(self):
        self.auth.addresses["test.domain"] = ["the.email.test.domain"]
        return self.mx.getMX("test.domain").addCallback(self._cbSimpleSuccess)

    def _cbSimpleSuccess(self, mx):
        self.assertEqual(mx.preference, 0)
        self.assertEqual(str(mx.name), "the.email.test.domain")

    def testSimpleFailure(self):
        self.mx.fallbackToDomain = False
        return self.assertFailure(self.mx.getMX("test.domain"), IOError)

    def testSimpleFailureWithFallback(self):
        return self.assertFailure(self.mx.getMX("test.domain"), DNSLookupError)

    def _exchangeTest(self, domain, records, correctMailExchange):
        """
        Issue an MX request for the given domain and arrange for it to be
        responded to with the given records.  Verify that the resulting mail
        exchange is the indicated host.

        @type domain: C{str}
        @type records: C{list} of L{RRHeader}
        @type correctMailExchange: C{str}
        @rtype: L{Deferred}
        """

        class DummyResolver:
            def lookupMailExchange(self, name):
                if name == domain:
                    return defer.succeed((records, [], []))
                return defer.fail(DNSNameError(domain))

        self.mx.resolver = DummyResolver()
        d = self.mx.getMX(domain)

        def gotMailExchange(record):
            self.assertEqual(str(record.name), correctMailExchange)

        d.addCallback(gotMailExchange)
        return d

    def test_mailExchangePreference(self):
        """
        The MX record with the lowest preference is returned by
        L{MXCalculator.getMX}.
        """
        domain = "example.com"
        good = "good.example.com"
        bad = "bad.example.com"

        records = [
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(1, bad)),
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(0, good)),
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(2, bad)),
        ]
        return self._exchangeTest(domain, records, good)

    def test_badExchangeExcluded(self):
        """
        L{MXCalculator.getMX} returns the MX record with the lowest preference
        which is not also marked as bad.
        """
        domain = "example.com"
        good = "good.example.com"
        bad = "bad.example.com"

        records = [
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(0, bad)),
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(1, good)),
        ]
        self.mx.markBad(bad)
        return self._exchangeTest(domain, records, good)

    def test_fallbackForAllBadExchanges(self):
        """
        L{MXCalculator.getMX} returns the MX record with the lowest preference
        if all the MX records in the response have been marked bad.
        """
        domain = "example.com"
        bad = "bad.example.com"
        worse = "worse.example.com"

        records = [
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(0, bad)),
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(1, worse)),
        ]
        self.mx.markBad(bad)
        self.mx.markBad(worse)
        return self._exchangeTest(domain, records, bad)

    def test_badExchangeExpires(self):
        """
        L{MXCalculator.getMX} returns the MX record with the lowest preference
        if it was last marked bad longer than L{MXCalculator.timeOutBadMX}
        seconds ago.
        """
        domain = "example.com"
        good = "good.example.com"
        previouslyBad = "bad.example.com"

        records = [
            RRHeader(
                name=domain, type=Record_MX.TYPE, payload=Record_MX(0, previouslyBad)
            ),
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(1, good)),
        ]
        self.mx.markBad(previouslyBad)
        self.clock.advance(self.mx.timeOutBadMX)
        return self._exchangeTest(domain, records, previouslyBad)

    def test_goodExchangeUsed(self):
        """
        L{MXCalculator.getMX} returns the MX record with the lowest preference
        if it was marked good after it was marked bad.
        """
        domain = "example.com"
        good = "good.example.com"
        previouslyBad = "bad.example.com"

        records = [
            RRHeader(
                name=domain, type=Record_MX.TYPE, payload=Record_MX(0, previouslyBad)
            ),
            RRHeader(name=domain, type=Record_MX.TYPE, payload=Record_MX(1, good)),
        ]
        self.mx.markBad(previouslyBad)
        self.mx.markGood(previouslyBad)
        self.clock.advance(self.mx.timeOutBadMX)
        return self._exchangeTest(domain, records, previouslyBad)

    def test_successWithoutResults(self):
        """
        If an MX lookup succeeds but the result set is empty,
        L{MXCalculator.getMX} should try to look up an I{A} record for the
        requested name and call back its returned Deferred with that
        address.
        """
        ip = "1.2.3.4"
        domain = "example.org"

        class DummyResolver:
            """
            Fake resolver which will respond to an MX lookup with an empty
            result set.

            @ivar mx: A dictionary mapping hostnames to three-tuples of
                results to be returned from I{MX} lookups.

            @ivar a: A dictionary mapping hostnames to addresses to be
                returned from I{A} lookups.
            """

            mx = {domain: ([], [], [])}
            a = {domain: ip}

            def lookupMailExchange(self, domain):
                return defer.succeed(self.mx[domain])

            def getHostByName(self, domain):
                return defer.succeed(self.a[domain])

        self.mx.resolver = DummyResolver()
        d = self.mx.getMX(domain)
        d.addCallback(self.assertEqual, Record_MX(name=ip))
        return d

    def test_failureWithSuccessfulFallback(self):
        """
        Test that if the MX record lookup fails, fallback is enabled, and an A
        record is available for the name, then the Deferred returned by
        L{MXCalculator.getMX} ultimately fires with a Record_MX instance which
        gives the address in the A record for the name.
        """

        class DummyResolver:
            """
            Fake resolver which will fail an MX lookup but then succeed a
            getHostByName call.
            """

            def lookupMailExchange(self, domain):
                return defer.fail(DNSNameError())

            def getHostByName(self, domain):
                return defer.succeed("1.2.3.4")

        self.mx.resolver = DummyResolver()
        d = self.mx.getMX("domain")
        d.addCallback(self.assertEqual, Record_MX(name="1.2.3.4"))
        return d

    def test_cnameWithoutGlueRecords(self):
        """
        If an MX lookup returns a single CNAME record as a result, MXCalculator
        will perform an MX lookup for the canonical name indicated and return
        the MX record which results.
        """
        alias = "alias.example.com"
        canonical = "canonical.example.com"
        exchange = "mail.example.com"

        class DummyResolver:
            """
            Fake resolver which will return a CNAME for an MX lookup of a name
            which is an alias and an MX for an MX lookup of the canonical name.
            """

            def lookupMailExchange(self, domain):
                if domain == alias:
                    return defer.succeed(
                        (
                            [
                                RRHeader(
                                    name=domain,
                                    type=Record_CNAME.TYPE,
                                    payload=Record_CNAME(canonical),
                                )
                            ],
                            [],
                            [],
                        )
                    )
                elif domain == canonical:
                    return defer.succeed(
                        (
                            [
                                RRHeader(
                                    name=domain,
                                    type=Record_MX.TYPE,
                                    payload=Record_MX(0, exchange),
                                )
                            ],
                            [],
                            [],
                        )
                    )
                else:
                    return defer.fail(DNSNameError(domain))

        self.mx.resolver = DummyResolver()
        d = self.mx.getMX(alias)
        d.addCallback(self.assertEqual, Record_MX(name=exchange))
        return d

    def test_cnameChain(self):
        """
        If L{MXCalculator.getMX} encounters a CNAME chain which is longer than
        the length specified, the returned L{Deferred} should errback with
        L{CanonicalNameChainTooLong}.
        """

        class DummyResolver:
            """
            Fake resolver which generates a CNAME chain of infinite length in
            response to MX lookups.
            """

            chainCounter = 0

            def lookupMailExchange(self, domain):
                self.chainCounter += 1
                name = "x-%d.example.com" % (self.chainCounter,)
                return defer.succeed(
                    (
                        [
                            RRHeader(
                                name=domain,
                                type=Record_CNAME.TYPE,
                                payload=Record_CNAME(name),
                            )
                        ],
                        [],
                        [],
                    )
                )

        cnameLimit = 3
        self.mx.resolver = DummyResolver()
        d = self.mx.getMX("mail.example.com", cnameLimit)
        self.assertFailure(d, twisted.mail.relaymanager.CanonicalNameChainTooLong)

        def cbChainTooLong(error):
            self.assertEqual(
                error.args[0], Record_CNAME("x-%d.example.com" % (cnameLimit + 1,))
            )
            self.assertEqual(self.mx.resolver.chainCounter, cnameLimit + 1)

        d.addCallback(cbChainTooLong)
        return d

    def test_cnameWithGlueRecords(self):
        """
        If an MX lookup returns a CNAME and the MX record for the CNAME, the
        L{Deferred} returned by L{MXCalculator.getMX} should be called back
        with the name from the MX record without further lookups being
        attempted.
        """
        lookedUp = []
        alias = "alias.example.com"
        canonical = "canonical.example.com"
        exchange = "mail.example.com"

        class DummyResolver:
            def lookupMailExchange(self, domain):
                if domain != alias or lookedUp:
                    # Don't give back any results for anything except the alias
                    # or on any request after the first.
                    return ([], [], [])
                return defer.succeed(
                    (
                        [
                            RRHeader(
                                name=alias,
                                type=Record_CNAME.TYPE,
                                payload=Record_CNAME(canonical),
                            ),
                            RRHeader(
                                name=canonical,
                                type=Record_MX.TYPE,
                                payload=Record_MX(name=exchange),
                            ),
                        ],
                        [],
                        [],
                    )
                )

        self.mx.resolver = DummyResolver()
        d = self.mx.getMX(alias)
        d.addCallback(self.assertEqual, Record_MX(name=exchange))
        return d

    def test_cnameLoopWithGlueRecords(self):
        """
        If an MX lookup returns two CNAME records which point to each other,
        the loop should be detected and the L{Deferred} returned by
        L{MXCalculator.getMX} should be errbacked with L{CanonicalNameLoop}.
        """
        firstAlias = "cname1.example.com"
        secondAlias = "cname2.example.com"

        class DummyResolver:
            def lookupMailExchange(self, domain):
                return defer.succeed(
                    (
                        [
                            RRHeader(
                                name=firstAlias,
                                type=Record_CNAME.TYPE,
                                payload=Record_CNAME(secondAlias),
                            ),
                            RRHeader(
                                name=secondAlias,
                                type=Record_CNAME.TYPE,
                                payload=Record_CNAME(firstAlias),
                            ),
                        ],
                        [],
                        [],
                    )
                )

        self.mx.resolver = DummyResolver()
        d = self.mx.getMX(firstAlias)
        self.assertFailure(d, twisted.mail.relaymanager.CanonicalNameLoop)
        return d

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testManyRecords(self):
        self.auth.addresses["test.domain"] = [
            "mx1.test.domain",
            "mx2.test.domain",
            "mx3.test.domain",
        ]
        return self.mx.getMX("test.domain").addCallback(
            self._cbManyRecordsSuccessfulLookup
        )

    def _cbManyRecordsSuccessfulLookup(self, mx):
        self.assertTrue(str(mx.name).split(".", 1)[0] in ("mx1", "mx2", "mx3"))
        self.mx.markBad(str(mx.name))
        return self.mx.getMX("test.domain").addCallback(
            self._cbManyRecordsDifferentResult, mx
        )

    def _cbManyRecordsDifferentResult(self, nextMX, mx):
        self.assertNotEqual(str(mx.name), str(nextMX.name))
        self.mx.markBad(str(nextMX.name))

        return self.mx.getMX("test.domain").addCallback(
            self._cbManyRecordsLastResult, mx, nextMX
        )

    def _cbManyRecordsLastResult(self, lastMX, mx, nextMX):
        self.assertNotEqual(str(mx.name), str(lastMX.name))
        self.assertNotEqual(str(nextMX.name), str(lastMX.name))

        self.mx.markBad(str(lastMX.name))
        self.mx.markGood(str(nextMX.name))

        return self.mx.getMX("test.domain").addCallback(
            self._cbManyRecordsRepeatSpecificResult, nextMX
        )

    def _cbManyRecordsRepeatSpecificResult(self, againMX, nextMX):
        self.assertEqual(str(againMX.name), str(nextMX.name))


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class LiveFireExerciseTests(TestCase):
    if interfaces.IReactorUDP(reactor, None) is None:
        skip = "UDP support is required to determining MX records"

    def setUp(self):
        setUpDNS(self)
        self.tmpdirs = [
            "domainDir",
            "insertionDomain",
            "insertionQueue",
            "destinationDomain",
            "destinationQueue",
        ]

    def tearDown(self):
        for d in self.tmpdirs:
            if os.path.exists(d):
                shutil.rmtree(d)
        return tearDownDNS(self)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testLocalDelivery(self):
        service = mail.mail.MailService()
        service.smtpPortal.registerChecker(cred.checkers.AllowAnonymousAccess())
        domain = mail.maildir.MaildirDirdbmDomain(service, "domainDir")
        domain.addUser("user", "password")
        service.addDomain("test.domain", domain)
        service.portals[""] = service.portals["test.domain"]
        map(service.portals[""].registerChecker, domain.getCredentialsCheckers())

        service.setQueue(mail.relay.DomainQueuer(service))

        f = service.getSMTPFactory()

        self.smtpServer = reactor.listenTCP(0, f, interface="127.0.0.1")

        client = LineSendingProtocol(
            [
                "HELO meson",
                "MAIL FROM: <user@hostname>",
                "RCPT TO: <user@test.domain>",
                "DATA",
                "This is the message",
                ".",
                "QUIT",
            ]
        )

        done = Deferred()
        f = protocol.ClientFactory()
        f.protocol = lambda: client
        f.clientConnectionLost = lambda *args: done.callback(None)
        reactor.connectTCP("127.0.0.1", self.smtpServer.getHost().port, f)

        def finished(ign):
            mbox = domain.requestAvatar("user", None, pop3.IMailbox)[1]
            msg = mbox.getMessage(0).read()
            self.assertNotEqual(msg.find("This is the message"), -1)

            return self.smtpServer.stopListening()

        done.addCallback(finished)
        return done

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testRelayDelivery(self):
        # Here is the service we will connect to and send mail from
        insServ = mail.mail.MailService()
        insServ.smtpPortal.registerChecker(cred.checkers.AllowAnonymousAccess())
        domain = mail.maildir.MaildirDirdbmDomain(insServ, "insertionDomain")
        insServ.addDomain("insertion.domain", domain)
        os.mkdir("insertionQueue")
        insServ.setQueue(mail.relaymanager.Queue("insertionQueue"))
        insServ.domains.setDefaultDomain(mail.relay.DomainQueuer(insServ))
        manager = mail.relaymanager.SmartHostSMTPRelayingManager(insServ.queue)
        manager.fArgs += ("test.identity.hostname",)
        helper = mail.relaymanager.RelayStateHelper(manager, 1)
        # Yoink!  Now the internet obeys OUR every whim!
        manager.mxcalc = mail.relaymanager.MXCalculator(self.resolver)
        # And this is our whim.
        self.auth.addresses["destination.domain"] = ["127.0.0.1"]

        f = insServ.getSMTPFactory()
        self.insServer = reactor.listenTCP(0, f, interface="127.0.0.1")

        # Here is the service the previous one will connect to for final
        # delivery
        destServ = mail.mail.MailService()
        destServ.smtpPortal.registerChecker(cred.checkers.AllowAnonymousAccess())
        domain = mail.maildir.MaildirDirdbmDomain(destServ, "destinationDomain")
        domain.addUser("user", "password")
        destServ.addDomain("destination.domain", domain)
        os.mkdir("destinationQueue")
        destServ.setQueue(mail.relaymanager.Queue("destinationQueue"))
        helper = mail.relaymanager.RelayStateHelper(manager, 1)
        helper.startService()

        f = destServ.getSMTPFactory()
        self.destServer = reactor.listenTCP(0, f, interface="127.0.0.1")

        # Update the port number the *first* relay will connect to, because we can't use
        # port 25
        manager.PORT = self.destServer.getHost().port

        client = LineSendingProtocol(
            [
                "HELO meson",
                "MAIL FROM: <user@wherever>",
                "RCPT TO: <user@destination.domain>",
                "DATA",
                "This is the message",
                ".",
                "QUIT",
            ]
        )

        done = Deferred()
        f = protocol.ClientFactory()
        f.protocol = lambda: client
        f.clientConnectionLost = lambda *args: done.callback(None)
        reactor.connectTCP("127.0.0.1", self.insServer.getHost().port, f)

        def finished(ign):
            # First part of the delivery is done.  Poke the queue manually now
            # so we don't have to wait for the queue to be flushed.
            delivery = manager.checkState()

            def delivered(ign):
                mbox = domain.requestAvatar("user", None, pop3.IMailbox)[1]
                msg = mbox.getMessage(0).read()
                self.assertNotEqual(msg.find("This is the message"), -1)

                self.insServer.stopListening()
                self.destServer.stopListening()
                helper.stopService()

            delivery.addCallback(delivered)
            return delivery

        done.addCallback(finished)
        return done


class LineBufferMessage:
    def __init__(self):
        self.lines = []
        self.eom = False
        self.lost = False

    def lineReceived(self, line):
        self.lines.append(line)

    def eomReceived(self):
        self.eom = True
        return defer.succeed("<Whatever>")

    def connectionLost(self):
        self.lost = True


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class AliasTests(TestCase):
    lines = ["First line", "Next line", "", "After a blank line", "Last line"]

    def testHandle(self):
        result = {}
        lines = [
            "user:  another@host\n",
            "nextuser:  |/bin/program\n",
            "user:  me@again\n",
            "moreusers: :/etc/include/filename\n",
            "multiuser: first@host, second@host,last@anotherhost",
        ]

        for l in lines:
            mail.alias.handle(result, l, "TestCase", None)

        self.assertEqual(result["user"], ["another@host", "me@again"])
        self.assertEqual(result["nextuser"], ["|/bin/program"])
        self.assertEqual(result["moreusers"], [":/etc/include/filename"])
        self.assertEqual(
            result["multiuser"], ["first@host", "second@host", "last@anotherhost"]
        )

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testFileLoader(self):
        domains = {"": object()}
        result = mail.alias.loadAliasFile(
            domains,
            fp=io.BytesIO(
                textwrap.dedent(
                    """\
                    # Here's a comment
                       # woop another one
                    testuser:                   address1,address2, address3,
                        continuation@address, |/bin/process/this

                    usertwo:thisaddress,thataddress, lastaddress
                    lastuser:       :/includable, /filename, |/program, address
                    """
                ).encode()
            ),
        )

        self.assertEqual(len(result), 3)

        group = result["testuser"]
        s = str(group)
        for a in (
            "address1",
            "address2",
            "address3",
            "continuation@address",
            "/bin/process/this",
        ):
            self.assertNotEqual(s.find(a), -1)
        self.assertEqual(len(group), 5)

        group = result["usertwo"]
        s = str(group)
        for a in ("thisaddress", "thataddress", "lastaddress"):
            self.assertNotEqual(s.find(a), -1)
        self.assertEqual(len(group), 3)

        group = result["lastuser"]
        s = str(group)
        self.assertEqual(s.find("/includable"), -1)
        for a in ("/filename", "program", "address"):
            self.assertNotEqual(s.find(a), -1, "%s not found" % a)
        self.assertEqual(len(group), 3)

    def testMultiWrapper(self):
        msgs = LineBufferMessage(), LineBufferMessage(), LineBufferMessage()
        msg = mail.alias.MultiWrapper(msgs)

        for L in self.lines:
            msg.lineReceived(L)
        return msg.eomReceived().addCallback(self._cbMultiWrapper, msgs)

    def _cbMultiWrapper(self, ignored, msgs):
        for m in msgs:
            self.assertTrue(m.eom)
            self.assertFalse(m.lost)
            self.assertEqual(self.lines, m.lines)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def testFileAlias(self):
        tmpfile = self.mktemp()
        a = mail.alias.FileAlias(tmpfile, None, None)
        m = a.createMessageReceiver()

        for l in self.lines:
            m.lineReceived(l)
        return m.eomReceived().addCallback(self._cbTestFileAlias, tmpfile)

    def _cbTestFileAlias(self, ignored, tmpfile):
        with open(tmpfile) as f:
            lines = f.readlines()
        self.assertEqual([L[:-1] for L in lines], self.lines)


class DummyDomain:
    """
    Test domain for L{AddressAliasTests}.
    """

    def __init__(self, address):
        self.address = address

    def exists(self, user, memo=None):
        """
        @returns: When a C{memo} is passed in this will raise a
            L{smtp.SMTPBadRcpt} exception, otherwise a boolean
            indicating if the C{user} and string version of
            L{self.address} are equal or not.
        @rtype: C{bool}
        """
        if memo:
            raise mail.smtp.SMTPBadRcpt("ham")

        return lambda: user == str(self.address)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class AddressAliasTests(TestCase):
    """
    Tests for L{twisted.mail.alias.AddressAlias}.
    """

    def setUp(self):
        """
        Setup an L{AddressAlias}.
        """
        self.address = mail.smtp.Address("foo@bar")
        domains = {self.address.domain: DummyDomain(self.address)}
        self.alias = mail.alias.AddressAlias(self.address, domains, self.address)

    def test_createMessageReceiver(self):
        """
        L{createMessageReceiever} calls C{exists()} on the domain object
        which key matches the C{alias} passed to L{AddressAlias}.
        """
        self.assertTrue(self.alias.createMessageReceiver())

    def test_str(self):
        """
        The string presentation of L{AddressAlias} includes the alias.
        """
        self.assertEqual(str(self.alias), "<Address foo@bar>")

    def test_resolve(self):
        """
        L{resolve} will look for additional aliases when an C{aliasmap}
        dictionary is passed, and returns L{None} if none were found.
        """
        self.assertEqual(self.alias.resolve({self.address: "bar"}), None)

    def test_resolveWithoutAliasmap(self):
        """
        L{resolve} returns L{None} when the alias could not be found in the
        C{aliasmap} and no L{mail.smtp.User} with this alias exists either.
        """
        self.assertEqual(self.alias.resolve({}), None)


class DummyProcess:
    __slots__ = ["onEnd"]


class MockProcessAlias(mail.alias.ProcessAlias):
    """
    An alias processor that doesn't actually launch processes.
    """

    def spawnProcess(self, proto, program, path):
        """
        Don't spawn a process.
        """


class MockAliasGroup(mail.alias.AliasGroup):
    """
    An alias group using C{MockProcessAlias}.
    """

    processAliasFactory = MockProcessAlias


class StubProcess:
    """
    Fake implementation of L{IProcessTransport}.

    @ivar signals: A list of all the signals which have been sent to this fake
        process.
    """

    def __init__(self):
        self.signals = []

    def loseConnection(self):
        """
        No-op implementation of disconnection.
        """

    def signalProcess(self, signal):
        """
        Record a signal sent to this process for later inspection.
        """
        self.signals.append(signal)


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class ProcessAliasTests(TestCase):
    """
    Tests for alias resolution.
    """

    if interfaces.IReactorProcess(reactor, None) is None:
        skip = "IReactorProcess not supported"

    lines = ["First line", "Next line", "", "After a blank line", "Last line"]

    def exitStatus(self, code):
        """
        Construct a status from the given exit code.

        @type code: L{int} between 0 and 255 inclusive.
        @param code: The exit status which the code will represent.

        @rtype: L{int}
        @return: A status integer for the given exit code.
        """
        # /* Macros for constructing status values.  */
        # #define __W_EXITCODE(ret, sig)  ((ret) << 8 | (sig))
        status = (code << 8) | 0

        # Sanity check
        self.assertTrue(os.WIFEXITED(status))
        self.assertEqual(os.WEXITSTATUS(status), code)
        self.assertFalse(os.WIFSIGNALED(status))

        return status

    def signalStatus(self, signal):
        """
        Construct a status from the given signal.

        @type signal: L{int} between 0 and 255 inclusive.
        @param signal: The signal number which the status will represent.

        @rtype: L{int}
        @return: A status integer for the given signal.
        """
        # /* If WIFSIGNALED(STATUS), the terminating signal.  */
        # #define __WTERMSIG(status)      ((status) & 0x7f)
        # /* Nonzero if STATUS indicates termination by a signal.  */
        # #define __WIFSIGNALED(status) \
        #    (((signed char) (((status) & 0x7f) + 1) >> 1) > 0)
        status = signal

        # Sanity check
        self.assertTrue(os.WIFSIGNALED(status))
        self.assertEqual(os.WTERMSIG(status), signal)
        self.assertFalse(os.WIFEXITED(status))

        return status

    def setUp(self):
        """
        Replace L{smtp.DNSNAME} with a well-known value.
        """
        self.DNSNAME = smtp.DNSNAME
        smtp.DNSNAME = ""

    def tearDown(self):
        """
        Restore the original value of L{smtp.DNSNAME}.
        """
        smtp.DNSNAME = self.DNSNAME

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_processAlias(self):
        """
        Standard call to C{mail.alias.ProcessAlias}: check that the specified
        script is called, and that the input is correctly transferred to it.
        """
        sh = FilePath(self.mktemp())
        sh.setContent(
            """\
#!/bin/sh
rm -f process.alias.out
while read i; do
    echo $i >> process.alias.out
done"""
        )
        os.chmod(sh.path, 0o700)
        a = mail.alias.ProcessAlias(sh.path, None, None)
        m = a.createMessageReceiver()

        for l in self.lines:
            m.lineReceived(l)

        def _cbProcessAlias(ignored):
            with open("process.alias.out") as f:
                lines = f.readlines()
            self.assertEqual([L[:-1] for L in lines], self.lines)

        return m.eomReceived().addCallback(_cbProcessAlias)

    def test_processAliasTimeout(self):
        """
        If the alias child process does not exit within a particular period of
        time, the L{Deferred} returned by L{MessageWrapper.eomReceived} should
        fail with L{ProcessAliasTimeout} and send the I{KILL} signal to the
        child process..
        """
        reactor = task.Clock()
        transport = StubProcess()
        proto = mail.alias.ProcessAliasProtocol()
        proto.makeConnection(transport)

        receiver = mail.alias.MessageWrapper(proto, None, reactor)
        d = receiver.eomReceived()
        reactor.advance(receiver.completionTimeout)

        def timedOut(ignored):
            self.assertEqual(transport.signals, ["KILL"])
            # Now that it has been killed, disconnect the protocol associated
            # with it.
            proto.processEnded(ProcessTerminated(self.signalStatus(signal.SIGKILL)))

        self.assertFailure(d, mail.alias.ProcessAliasTimeout)
        d.addCallback(timedOut)
        return d

    def test_earlyProcessTermination(self):
        """
        If the process associated with an L{mail.alias.MessageWrapper} exits
        before I{eomReceived} is called, the L{Deferred} returned by
        I{eomReceived} should fail.
        """
        transport = StubProcess()
        protocol = mail.alias.ProcessAliasProtocol()
        protocol.makeConnection(transport)
        receiver = mail.alias.MessageWrapper(protocol, None, None)
        protocol.processEnded(failure.Failure(ProcessDone(0)))
        return self.assertFailure(receiver.eomReceived(), ProcessDone)

    def _terminationTest(self, status):
        """
        Verify that if the process associated with an
        L{mail.alias.MessageWrapper} exits with the given status, the
        L{Deferred} returned by I{eomReceived} fails with L{ProcessTerminated}.
        """
        transport = StubProcess()
        protocol = mail.alias.ProcessAliasProtocol()
        protocol.makeConnection(transport)
        receiver = mail.alias.MessageWrapper(protocol, None, None)
        protocol.processEnded(failure.Failure(ProcessTerminated(status)))
        return self.assertFailure(receiver.eomReceived(), ProcessTerminated)

    def test_errorProcessTermination(self):
        """
        If the process associated with an L{mail.alias.MessageWrapper} exits
        with a non-zero exit code, the L{Deferred} returned by I{eomReceived}
        should fail.
        """
        return self._terminationTest(self.exitStatus(1))

    def test_signalProcessTermination(self):
        """
        If the process associated with an L{mail.alias.MessageWrapper} exits
        because it received a signal, the L{Deferred} returned by
        I{eomReceived} should fail.
        """
        return self._terminationTest(self.signalStatus(signal.SIGHUP))

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_aliasResolution(self):
        """
        Check that the C{resolve} method of alias processors produce the correct
        set of objects:
            - direct alias with L{mail.alias.AddressAlias} if a simple input is passed
            - aliases in a file with L{mail.alias.FileWrapper} if an input in the format
              '/file' is given
            - aliases resulting of a process call wrapped by L{mail.alias.MessageWrapper}
              if the format is '|process'
        """
        aliases = {}
        domain = {"": TestDomain(aliases, ["user1", "user2", "user3"])}
        A1 = MockAliasGroup(["user1", "|echo", "/file"], domain, "alias1")
        A2 = MockAliasGroup(["user2", "user3"], domain, "alias2")
        A3 = mail.alias.AddressAlias("alias1", domain, "alias3")
        aliases.update(
            {
                "alias1": A1,
                "alias2": A2,
                "alias3": A3,
            }
        )

        res1 = A1.resolve(aliases)
        r1 = map(str, res1.objs)
        r1.sort()
        expected = map(
            str,
            [
                mail.alias.AddressAlias("user1", None, None),
                mail.alias.MessageWrapper(DummyProcess(), "echo"),
                mail.alias.FileWrapper("/file"),
            ],
        )
        expected.sort()
        self.assertEqual(r1, expected)

        res2 = A2.resolve(aliases)
        r2 = map(str, res2.objs)
        r2.sort()
        expected = map(
            str,
            [
                mail.alias.AddressAlias("user2", None, None),
                mail.alias.AddressAlias("user3", None, None),
            ],
        )
        expected.sort()
        self.assertEqual(r2, expected)

        res3 = A3.resolve(aliases)
        r3 = map(str, res3.objs)
        r3.sort()
        expected = map(
            str,
            [
                mail.alias.AddressAlias("user1", None, None),
                mail.alias.MessageWrapper(DummyProcess(), "echo"),
                mail.alias.FileWrapper("/file"),
            ],
        )
        expected.sort()
        self.assertEqual(r3, expected)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_cyclicAlias(self):
        """
        Check that a cycle in alias resolution is correctly handled.
        """
        aliases = {}
        domain = {"": TestDomain(aliases, [])}
        A1 = mail.alias.AddressAlias("alias2", domain, "alias1")
        A2 = mail.alias.AddressAlias("alias3", domain, "alias2")
        A3 = mail.alias.AddressAlias("alias1", domain, "alias3")
        aliases.update({"alias1": A1, "alias2": A2, "alias3": A3})

        self.assertEqual(aliases["alias1"].resolve(aliases), None)
        self.assertEqual(aliases["alias2"].resolve(aliases), None)
        self.assertEqual(aliases["alias3"].resolve(aliases), None)

        A4 = MockAliasGroup(["|echo", "alias1"], domain, "alias4")
        aliases["alias4"] = A4

        res = A4.resolve(aliases)
        r = map(str, res.objs)
        r.sort()
        expected = map(str, [mail.alias.MessageWrapper(DummyProcess(), "echo")])
        expected.sort()
        self.assertEqual(r, expected)


class TestDomain:
    def __init__(self, aliases, users):
        self.aliases = aliases
        self.users = users

    def exists(self, user, memo=None):
        user = user.dest.local
        if user in self.users:
            return lambda: mail.alias.AddressAlias(user, None, None)
        try:
            a = self.aliases[user]
        except BaseException:
            raise smtp.SMTPBadRcpt(user)
        else:
            aliases = a.resolve(self.aliases, memo)
            if aliases:
                return lambda: aliases
            raise smtp.SMTPBadRcpt(user)


class DummyQueue:
    """
    A fake relay queue to use for testing.

    This queue doesn't keep track of which messages are waiting to be relayed
    or are in the process of being relayed.

    @ivar directory: See L{__init__}.
    """

    def __init__(self, directory):
        """
        @type directory: L{bytes}
        @param directory: The pathname of the directory holding messages in the
            queue.
        """
        self.directory = directory

    def done(self, message):
        """
        Remove a message from the queue.

        @type message: L{bytes}
        @param message: The base filename of a message.
        """
        message = os.path.basename(message)
        os.remove(self.getPath(message) + "-D")
        os.remove(self.getPath(message) + "-H")

    def getEnvelopeFile(self, message):
        """
        Get the envelope file for a message in the queue.

        @type message: L{bytes}
        @param message: The base filename of a message.

        @rtype: L{file}
        @return: The envelope file for the message.
        """
        return open(os.path.join(self.directory, message + "-H"), "rb")

    def getPath(self, message):
        """
        Return the full base pathname of a message in the queue.

        @type message: L{bytes}
        @param message: The base filename of a message.

        @rtype: L{bytes}
        @return: The full base pathname of the message.
        """
        return os.path.join(self.directory, message)

    def createNewMessage(self):
        """
        Create a new message in the queue.

        @rtype: 2-L{tuple} of (E{1}) L{file}, (E{2}) L{FileMessage}
        @return: The envelope file and a message receiver for a new message in
            the queue.
        """
        fname = f"{time.time()}_{id(self)}"
        headerFile = open(os.path.join(self.directory, fname + "-H"), "wb")
        tempFilename = os.path.join(self.directory, fname + "-C")
        finalFilename = os.path.join(self.directory, fname + "-D")
        messageFile = open(tempFilename, "wb")

        return headerFile, mail.mail.FileMessage(
            messageFile, tempFilename, finalFilename
        )

    def setWaiting(self, message):
        """
        Ignore the request to mark a message as waiting to be relayed.

        @type message: L{bytes}
        @param message: The base filename of a message.
        """
        pass


class DummySmartHostSMTPRelayingManager:
    """
    A fake smart host to use for testing.

    @type managed: L{dict} of L{bytes} -> L{list} of
        L{list} of L{bytes}
    @ivar managed: A mapping of a string identifying a managed relayer to
        filenames of messages the managed relayer is responsible for.

    @ivar queue: See L{__init__}.
    """

    def __init__(self, queue):
        """
        Initialize the minimum necessary members of a smart host.

        @type queue: L{DummyQueue}
        @param queue: A queue that can be used for testing purposes.
        """
        self.managed = {}
        self.queue = queue


@skipIf(platformType != "posix", "twisted.mail only works on posix")
class _AttemptManagerTests(TestCase):
    """
    Test the behavior of L{_AttemptManager}.

    @type tmpdir: L{bytes}
    @ivar tmpdir: The path to a temporary directory holding the message files.

    @type reactor: L{MemoryReactorClock}
    @ivar reactor: The reactor used for test purposes.

    @type eventLog: L{None} or L{dict} of L{bytes} -> L{object}
    @ivar eventLog: Information about the last informational log message
        generated or none if no log message has been generated.

    @type noisyAttemptMgr: L{_AttemptManager}
    @ivar noisyAttemptMgr: An attempt manager which generates informational
        log messages.

    @type quietAttemptMgr: L{_AttemptManager}
    @ivar quietAttemptMgr: An attempt manager which does not generate
        informational log messages.

    @type noisyMessage: L{bytes}
    @ivar noisyMessage: The full base pathname of the message to be used with
        the noisy attempt manager.

    @type quietMessage: L{bytes}
    @ivar quietMessage: The full base pathname of the message to be used with
        the quiet.
    """

    def setUp(self):
        """
        Set up a temporary directory for the queue, attempt managers with the
        noisy flag on and off, message files for use with each attempt manager,
        and a reactor.  Also, register to be notified when log messages are
        generated.
        """
        self.tmpdir = self.mktemp()
        os.mkdir(self.tmpdir)

        self.reactor = MemoryReactorClock()

        self.eventLog = None
        log.addObserver(self._logObserver)

        self.noisyAttemptMgr = _AttemptManager(
            DummySmartHostSMTPRelayingManager(DummyQueue(self.tmpdir)),
            True,
            self.reactor,
        )
        self.quietAttemptMgr = _AttemptManager(
            DummySmartHostSMTPRelayingManager(DummyQueue(self.tmpdir)),
            False,
            self.reactor,
        )

        noisyBaseName = "noisyMessage"
        quietBaseName = "quietMessage"

        self.noisyMessage = os.path.join(self.tmpdir, noisyBaseName)
        self.quietMessage = os.path.join(self.tmpdir, quietBaseName)

        open(self.noisyMessage + "-D", "w").close()

        open(self.quietMessage + "-D", "w").close()

        self.noisyAttemptMgr.manager.managed["noisyRelayer"] = [noisyBaseName]
        self.quietAttemptMgr.manager.managed["quietRelayer"] = [quietBaseName]

        with open(self.noisyMessage + "-H", "wb") as envelope:
            pickle.dump(["from-noisy@domain", "to-noisy@domain"], envelope)

        with open(self.quietMessage + "-H", "wb") as envelope:
            pickle.dump(["from-quiet@domain", "to-quiet@domain"], envelope)

    def tearDown(self):
        """
        Unregister for log events and remove the temporary directory.
        """
        log.removeObserver(self._logObserver)
        shutil.rmtree(self.tmpdir)

    def _logObserver(self, eventDict):
        """
        A log observer.

        @type eventDict: L{dict} of L{bytes} -> L{object}
        @param eventDict: Information about the last informational log message
            generated.
        """
        self.eventLog = eventDict

    def test_initNoisyDefault(self):
        """
        When an attempt manager is created without the noisy parameter, the
        noisy instance variable should default to true.
        """
        am = _AttemptManager(DummySmartHostSMTPRelayingManager(DummyQueue(self.tmpdir)))
        self.assertTrue(am.noisy)

    def test_initNoisy(self):
        """
        When an attempt manager is created with the noisy parameter set to
        true, the noisy instance variable should be set to true.
        """
        self.assertTrue(self.noisyAttemptMgr.noisy)

    def test_initQuiet(self):
        """
        When an attempt manager is created with the noisy parameter set to
        false, the noisy instance variable should be set to false.
        """
        self.assertFalse(self.quietAttemptMgr.noisy)

    def test_initReactorDefault(self):
        """
        When an attempt manager is created without the reactor parameter, the
        reactor instance variable should default to the global reactor.
        """
        am = _AttemptManager(DummySmartHostSMTPRelayingManager(DummyQueue(self.tmpdir)))
        self.assertEqual(am.reactor, reactor)

    def test_initReactor(self):
        """
        When an attempt manager is created with a reactor provided, the
        reactor instance variable should default to that reactor.
        """
        self.assertEqual(self.noisyAttemptMgr.reactor, self.reactor)

    def test_notifySuccessNoisy(self):
        """
        For an attempt manager with the noisy flag set, notifySuccess should
        result in a log message.
        """
        self.noisyAttemptMgr.notifySuccess("noisyRelayer", self.noisyMessage)
        self.assertTrue(self.eventLog)

    def test_notifySuccessQuiet(self):
        """
        For an attempt manager with the noisy flag not set, notifySuccess
        should result in no log message.
        """
        self.quietAttemptMgr.notifySuccess("quietRelayer", self.quietMessage)
        self.assertFalse(self.eventLog)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_notifyFailureNoisy(self):
        """
        For an attempt manager with the noisy flag set, notifyFailure should
        result in a log message.
        """
        self.noisyAttemptMgr.notifyFailure("noisyRelayer", self.noisyMessage)
        self.assertTrue(self.eventLog)

    @skipIf(sys.version_info >= (3,), "not ported to Python 3")
    def test_notifyFailureQuiet(self):
        """
        For an attempt manager with the noisy flag not set, notifyFailure
        should result in no log message.
        """
        self.quietAttemptMgr.notifyFailure("quietRelayer", self.quietMessage)
        self.assertFalse(self.eventLog)

    def test_notifyDoneNoisy(self):
        """
        For an attempt manager with the noisy flag set, notifyDone should
        result in a log message.
        """
        self.noisyAttemptMgr.notifyDone("noisyRelayer")
        self.assertTrue(self.eventLog)

    def test_notifyDoneQuiet(self):
        """
        For an attempt manager with the noisy flag not set, notifyDone
        should result in no log message.
        """
        self.quietAttemptMgr.notifyDone("quietRelayer")
        self.assertFalse(self.eventLog)

    def test_notifyNoConnectionNoisy(self):
        """
        For an attempt manager with the noisy flag set, notifyNoConnection
        should result in a log message.
        """
        self.noisyAttemptMgr.notifyNoConnection("noisyRelayer")
        self.assertTrue(self.eventLog)
        self.reactor.advance(60)

    def test_notifyNoConnectionQuiet(self):
        """
        For an attempt manager with the noisy flag not set, notifyNoConnection
        should result in no log message.
        """
        self.quietAttemptMgr.notifyNoConnection("quietRelayer")
        self.assertFalse(self.eventLog)
        self.reactor.advance(60)



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Support module for making a port forwarder with twistd.
"""
from twisted.application import strports
from twisted.protocols import portforward
from twisted.python import usage


class Options(usage.Options):
    synopsis = "[options]"
    longdesc = "(Deprecated) Port Forwarder. See 'twist forward' instead."
    optParameters = [
        ["port", "p", "tcp:6666", "The string endpoint to listen on."],
        ["host", "h", "localhost", "The destination host to connect to."],
        ["dest_port", "d", 6665, "Set the destination port to connect to."],
    ]

    compData = usage.Completions(optActions={"host": usage.CompleteHostnames()})


def makeService(config):
    """
    Create a port-forwarding service.
    """
    f = portforward.ProxyFactory(config["host"], int(config["dest_port"]))
    return strports.service(config["port"], f)



# -*- test-case-name: twisted.test.test_usage -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.


"""
twisted.python.usage is a module for parsing/handling the
command line of your program.

For information on how to use it, see
U{http://twistedmatrix.com/projects/core/documentation/howto/options.html},
or doc/core/howto/options.xhtml in your Twisted directory.
"""
from __future__ import annotations

import getopt

# System Imports
import inspect
import os
import sys
import textwrap
from os import path
from typing import Any, cast

# Sibling Imports
from twisted.python import reflect, util


class UsageError(Exception):
    pass


error = UsageError


class CoerceParameter:
    """
    Utility class that can corce a parameter before storing it.
    """

    def __init__(self, options, coerce):
        """
        @param options: parent Options object
        @param coerce: callable used to coerce the value.
        """
        self.options = options
        self.coerce = coerce
        self.doc = getattr(self.coerce, "coerceDoc", "")

    def dispatch(self, parameterName, value):
        """
        When called in dispatch, do the coerce for C{value} and save the
        returned value.
        """
        if value is None:
            raise UsageError(f"Parameter '{parameterName}' requires an argument.")
        try:
            value = self.coerce(value)
        except ValueError as e:
            raise UsageError(f"Parameter type enforcement failed: {e}")

        self.options.opts[parameterName] = value


class Options(dict[str, Any]):
    """
    An option list parser class

    C{optFlags} and C{optParameters} are lists of available parameters
    which your program can handle. The difference between the two
    is the 'flags' have an on(1) or off(0) state (off by default)
    whereas 'parameters' have an assigned value, with an optional
    default. (Compare '--verbose' and '--verbosity=2')

    optFlags is assigned a list of lists. Each list represents
    a flag parameter, as so::

       optFlags = [['verbose', 'v', 'Makes it tell you what it doing.'],
                   ['quiet', 'q', 'Be vewy vewy quiet.']]

    As you can see, the first item is the long option name
    (prefixed with '--' on the command line), followed by the
    short option name (prefixed with '-'), and the description.
    The description is used for the built-in handling of the
    --help switch, which prints a usage summary.

    C{optParameters} is much the same, except the list also contains
    a default value::

       optParameters = [['outfile', 'O', 'outfile.log', 'Description...']]

    A coerce function can also be specified as the last element: it will be
    called with the argument and should return the value that will be stored
    for the option. This function can have a C{coerceDoc} attribute which
    will be appended to the documentation of the option.

    subCommands is a list of 4-tuples of (command name, command shortcut,
    parser class, documentation).  If the first non-option argument found is
    one of the given command names, an instance of the given parser class is
    instantiated and given the remainder of the arguments to parse and
    self.opts[command] is set to the command name.  For example::

       subCommands = [
            ['inquisition', 'inquest', InquisitionOptions,
                 'Perform an inquisition'],
            ['holyquest', 'quest', HolyQuestOptions,
                 'Embark upon a holy quest']
        ]

    In this case, C{"<program> holyquest --horseback --for-grail"} will cause
    C{HolyQuestOptions} to be instantiated and asked to parse
    C{['--horseback', '--for-grail']}.  Currently, only the first sub-command
    is parsed, and all options following it are passed to its parser.  If a
    subcommand is found, the subCommand attribute is set to its name and the
    subOptions attribute is set to the Option instance that parses the
    remaining options. If a subcommand is not given to parseOptions,
    the subCommand attribute will be None. You can also mark one of
    the subCommands to be the default::

       defaultSubCommand = 'holyquest'

    In this case, the subCommand attribute will never be None, and
    the subOptions attribute will always be set.

    If you want to handle your own options, define a method named
    C{opt_paramname} that takes C{(self, option)} as arguments. C{option}
    will be whatever immediately follows the parameter on the
    command line. Options fully supports the mapping interface, so you
    can do things like C{'self["option"] = val'} in these methods.

    Shell tab-completion is supported by this class, for zsh only at present.
    Zsh ships with a stub file ("completion function") which, for Twisted
    commands, performs tab-completion on-the-fly using the support provided
    by this class. The stub file lives in our tree at
    C{twisted/python/twisted-completion.zsh}, and in the Zsh tree at
    C{Completion/Unix/Command/_twisted}.

    Tab-completion is based upon the contents of the optFlags and optParameters
    lists. And, optionally, additional metadata may be provided by assigning a
    special attribute, C{compData}, which should be an instance of
    C{Completions}. See that class for details of what can and should be
    included - and see the howto for additional help using these features -
    including how third-parties may take advantage of tab-completion for their
    own commands.

    Advanced functionality is covered in the howto documentation,
    available at
    U{http://twistedmatrix.com/projects/core/documentation/howto/options.html},
    or doc/core/howto/options.xhtml in your Twisted directory.
    """

    subCommand: str | None = None
    defaultSubCommand: str | None = None
    parent: Options | None = None
    completionData = None
    _shellCompFile = sys.stdout  # file to use if shell completion is requested

    def __init__(self):
        super().__init__()

        self.opts = self
        self.defaults = {}

        # These are strings/lists we will pass to getopt
        self.longOpt = []
        self.shortOpt = ""
        self.docs = {}
        self.synonyms = {}
        self._dispatch = {}

        collectors = [
            self._gather_flags,
            self._gather_parameters,
            self._gather_handlers,
        ]

        for c in collectors:
            (longOpt, shortOpt, docs, settings, synonyms, dispatch) = c()
            self.longOpt.extend(longOpt)
            self.shortOpt = self.shortOpt + shortOpt
            self.docs.update(docs)

            self.opts.update(settings)
            self.defaults.update(settings)

            self.synonyms.update(synonyms)
            self._dispatch.update(dispatch)

    # class Options derives from dict, which defines __hash__ as None,
    # but we need to set __hash__ to object.__hash__ which is of type
    # Callable[[object], int].  So we need to ignore mypy error here.
    __hash__ = object.__hash__  # type: ignore[assignment]

    def opt_help(self):
        """
        Display this help and exit.
        """
        print(self.__str__())
        sys.exit(0)

    def opt_version(self):
        """
        Display Twisted version and exit.
        """
        from twisted import copyright

        print("Twisted version:", copyright.version)
        sys.exit(0)

    # opt_h = opt_help # this conflicted with existing 'host' options.

    def parseOptions(self, options=None):
        """
        The guts of the command-line parser.
        """

        if options is None:
            options = sys.argv[1:]

        # we really do need to place the shell completion check here, because
        # if we used an opt_shell_completion method then it would be possible
        # for other opt_* methods to be run first, and they could possibly
        # raise validation errors which would result in error output on the
        # terminal of the user performing shell completion. Validation errors
        # would occur quite frequently, in fact, because users often initiate
        # tab-completion while they are editing an unfinished command-line.
        if len(options) > 1 and options[-2] == "--_shell-completion":
            from twisted.python import _shellcomp

            cmdName = path.basename(sys.argv[0])
            _shellcomp.shellComplete(self, cmdName, options, self._shellCompFile)
            sys.exit(0)

        try:
            opts, args = getopt.getopt(options, self.shortOpt, self.longOpt)
        except getopt.error as e:
            raise UsageError(str(e))

        for opt, arg in opts:
            if opt[1] == "-":
                opt = opt[2:]
            else:
                opt = opt[1:]

            optMangled = opt
            if optMangled not in self.synonyms:
                optMangled = opt.replace("-", "_")
                if optMangled not in self.synonyms:
                    raise UsageError(f"No such option '{opt}'")

            optMangled = self.synonyms[optMangled]
            if isinstance(self._dispatch[optMangled], CoerceParameter):
                self._dispatch[optMangled].dispatch(optMangled, arg)
            else:
                self._dispatch[optMangled](optMangled, arg)

        if getattr(self, "subCommands", None) and (
            args or self.defaultSubCommand is not None
        ):
            if not args:
                args = [self.defaultSubCommand]
            sub, rest = args[0], args[1:]
            for cmd, short, parser, doc in self.subCommands:
                if sub == cmd or sub == short:
                    self.subCommand = cmd
                    self.subOptions = parser()
                    self.subOptions.parent = self
                    self.subOptions.parseOptions(rest)
                    break
            else:
                raise UsageError("Unknown command: %s" % sub)
        else:
            try:
                self.parseArgs(*args)
            except TypeError:
                raise UsageError("Wrong number of arguments.")

        self.postOptions()

    def postOptions(self):
        """
        I am called after the options are parsed.

        Override this method in your subclass to do something after
        the options have been parsed and assigned, like validate that
        all options are sane.
        """

    def parseArgs(self):
        """
        I am called with any leftover arguments which were not options.

        Override me to do something with the remaining arguments on
        the command line, those which were not flags or options. e.g.
        interpret them as a list of files to operate on.

        Note that if there more arguments on the command line
        than this method accepts, parseArgs will blow up with
        a getopt.error.  This means if you don't override me,
        parseArgs will blow up if I am passed any arguments at
        all!
        """

    def _generic_flag(self, flagName, value=None):
        if value not in ("", None):
            raise UsageError(
                "Flag '%s' takes no argument." ' Not even "%s".' % (flagName, value)
            )

        self.opts[flagName] = 1

    def _gather_flags(self):
        """
        Gather up boolean (flag) options.
        """

        longOpt, shortOpt = [], ""
        docs, settings, synonyms, dispatch = {}, {}, {}, {}

        flags = []
        reflect.accumulateClassList(self.__class__, "optFlags", flags)

        for flag in flags:
            long, short, doc = util.padTo(3, flag)
            if not long:
                raise ValueError("A flag cannot be without a name.")

            docs[long] = doc
            settings[long] = 0
            if short:
                shortOpt = shortOpt + short
                synonyms[short] = long
            longOpt.append(long)
            synonyms[long] = long
            dispatch[long] = self._generic_flag

        return longOpt, shortOpt, docs, settings, synonyms, dispatch

    def _gather_parameters(self):
        """
        Gather options which take a value.
        """
        longOpt, shortOpt = [], ""
        docs, settings, synonyms, dispatch = {}, {}, {}, {}

        parameters = []

        reflect.accumulateClassList(self.__class__, "optParameters", parameters)

        synonyms = {}

        for parameter in parameters:
            long, short, default, doc, paramType = util.padTo(5, parameter)
            if not long:
                raise ValueError("A parameter cannot be without a name.")

            docs[long] = doc
            settings[long] = default
            if short:
                shortOpt = shortOpt + short + ":"
                synonyms[short] = long
            longOpt.append(long + "=")
            synonyms[long] = long
            if paramType is not None:
                dispatch[long] = CoerceParameter(self, paramType)
            else:
                dispatch[long] = CoerceParameter(self, str)

        return longOpt, shortOpt, docs, settings, synonyms, dispatch

    def _gather_handlers(self):
        """
        Gather up options with their own handler methods.

        This returns a tuple of many values.  Amongst those values is a
        synonyms dictionary, mapping all of the possible aliases (C{str})
        for an option to the longest spelling of that option's name
        C({str}).

        Another element is a dispatch dictionary, mapping each user-facing
        option name (with - substituted for _) to a callable to handle that
        option.
        """

        longOpt, shortOpt = [], ""
        docs, settings, synonyms, dispatch = {}, {}, {}, {}

        dct = {}
        reflect.addMethodNamesToDict(self.__class__, dct, "opt_")

        for name in dct.keys():
            method = getattr(self, "opt_" + name)

            takesArg = not flagFunction(method, name)

            prettyName = name.replace("_", "-")
            doc = getattr(method, "__doc__", None)
            if doc:
                ## Only use the first line.
                # docs[name] = doc.split('\n')[0]
                docs[prettyName] = doc
            else:
                docs[prettyName] = self.docs.get(prettyName)

            synonyms[prettyName] = prettyName

            # A little slight-of-hand here makes dispatching much easier
            # in parseOptions, as it makes all option-methods have the
            # same signature.
            if takesArg:
                fn = lambda name, value, m=method: m(value)
            else:
                # XXX: This won't raise a TypeError if it's called
                # with a value when it shouldn't be.
                fn = lambda name, value=None, m=method: m()

            dispatch[prettyName] = fn

            if len(name) == 1:
                shortOpt = shortOpt + name
                if takesArg:
                    shortOpt = shortOpt + ":"
            else:
                if takesArg:
                    prettyName = prettyName + "="
                longOpt.append(prettyName)

        reverse_dct = {}
        # Map synonyms
        for name in dct.keys():
            method = getattr(self, "opt_" + name)
            if method not in reverse_dct:
                reverse_dct[method] = []
            reverse_dct[method].append(name.replace("_", "-"))

        for method, names in reverse_dct.items():
            if len(names) < 2:
                continue
            longest = max(names, key=len)
            for name in names:
                synonyms[name] = longest

        return longOpt, shortOpt, docs, settings, synonyms, dispatch

    def __str__(self) -> str:
        return self.getSynopsis() + "\n" + self.getUsage(width=None)

    def getSynopsis(self) -> str:
        """
        Returns a string containing a description of these options and how to
        pass them to the executed file.
        """
        executableName = reflect.filenameToModuleName(sys.argv[0])

        if executableName.endswith(".__main__"):
            executableName = "{} -m {}".format(
                os.path.basename(sys.executable),
                executableName.replace(".__main__", ""),
            )

        if self.parent is None:
            default = "Usage: {}{}".format(
                executableName,
                (self.longOpt and " [options]") or "",
            )
        else:
            default = "%s" % ((self.longOpt and "[options]") or "")
        synopsis = cast(str, getattr(self, "synopsis", default))

        synopsis = synopsis.rstrip()

        if self.parent is not None:
            assert self.parent.subCommand is not None
            synopsis = " ".join(
                (self.parent.getSynopsis(), self.parent.subCommand, synopsis)
            )
        return synopsis

    def getUsage(self, width: int | None = None) -> str:
        # If subOptions exists by now, then there was probably an error while
        # parsing its options.
        if hasattr(self, "subOptions"):
            return cast(Options, self.subOptions).getUsage(width=width)

        if not width:
            width = int(os.environ.get("COLUMNS", "80"))

        if hasattr(self, "subCommands"):
            cmdDicts = []
            for cmd, short, parser, desc in self.subCommands:
                cmdDicts.append(
                    {
                        "long": cmd,
                        "short": short,
                        "doc": desc,
                        "optType": "command",
                        "default": None,
                    }
                )
            chunks = docMakeChunks(cmdDicts, width)
            commands = "Commands:\n" + "".join(chunks)
        else:
            commands = ""

        longToShort = {}
        for key, value in self.synonyms.items():
            longname = value
            if (key != longname) and (len(key) == 1):
                longToShort[longname] = key
            else:
                if longname not in longToShort:
                    longToShort[longname] = None
                else:
                    pass

        optDicts = []
        for opt in self.longOpt:
            if opt[-1] == "=":
                optType = "parameter"
                opt = opt[:-1]
            else:
                optType = "flag"

            optDicts.append(
                {
                    "long": opt,
                    "short": longToShort[opt],
                    "doc": self.docs[opt],
                    "optType": optType,
                    "default": self.defaults.get(opt, None),
                    "dispatch": self._dispatch.get(opt, None),
                }
            )

        if not (getattr(self, "longdesc", None) is None):
            longdesc = cast(str, self.longdesc)  # type: ignore[attr-defined]
        else:
            import __main__

            if getattr(__main__, "__doc__", None):
                longdesc = __main__.__doc__
            else:
                longdesc = ""

        if longdesc:
            longdesc = "\n" + "\n".join(textwrap.wrap(longdesc, width)).strip() + "\n"

        if optDicts:
            chunks = docMakeChunks(optDicts, width)
            s = "Options:\n%s" % ("".join(chunks))
        else:
            s = "Options: None\n"

        return s + longdesc + commands


_ZSH = "zsh"
_BASH = "bash"


class Completer:
    """
    A completion "action" - provides completion possibilities for a particular
    command-line option. For example we might provide the user a fixed list of
    choices, or files/dirs according to a glob.

    This class produces no completion matches itself - see the various
    subclasses for specific completion functionality.
    """

    _descr: str | None = None

    def __init__(self, descr=None, repeat=False):
        """
        @type descr: C{str}
        @param descr: An optional descriptive string displayed above matches.

        @type repeat: C{bool}
        @param repeat: A flag, defaulting to False, indicating whether this
            C{Completer} should repeat - that is, be used to complete more
            than one command-line word. This may ONLY be set to True for
            actions in the C{extraActions} keyword argument to C{Completions}.
            And ONLY if it is the LAST (or only) action in the C{extraActions}
            list.
        """
        if descr is not None:
            self._descr = descr
        self._repeat = repeat

    @property
    def _repeatFlag(self):
        if self._repeat:
            return "*"
        else:
            return ""

    def _description(self, optName):
        if self._descr is not None:
            return self._descr
        else:
            return optName

    def _shellCode(self, optName, shellType):
        """
        Fetch a fragment of shell code representing this action which is
        suitable for use by the completion system in _shellcomp.py

        @type optName: C{str}
        @param optName: The long name of the option this action is being
            used for.

        @type shellType: C{str}
        @param shellType: One of the supported shell constants e.g.
            C{twisted.python.usage._ZSH}
        """
        if shellType == _ZSH:
            return f"{self._repeatFlag}:{self._description(optName)}:"
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteFiles(Completer):
    """
    Completes file names based on a glob pattern
    """

    def __init__(self, globPattern="*", **kw):
        Completer.__init__(self, **kw)
        self._globPattern = globPattern

    def _description(self, optName):
        if self._descr is not None:
            return f"{self._descr} ({self._globPattern})"
        else:
            return f"{optName} ({self._globPattern})"

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            return '{}:{}:_files -g "{}"'.format(
                self._repeatFlag,
                self._description(optName),
                self._globPattern,
            )
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteDirs(Completer):
    """
    Completes directory names
    """

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            return "{}:{}:_directories".format(
                self._repeatFlag, self._description(optName)
            )
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteList(Completer):
    """
    Completes based on a fixed list of words
    """

    def __init__(self, items, **kw):
        Completer.__init__(self, **kw)
        self._items = items

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            return "{}:{}:({})".format(
                self._repeatFlag,
                self._description(optName),
                " ".join(
                    self._items,
                ),
            )
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteMultiList(Completer):
    """
    Completes multiple comma-separated items based on a fixed list of words
    """

    def __init__(self, items, **kw):
        Completer.__init__(self, **kw)
        self._items = items

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            return "{}:{}:_values -s , '{}' {}".format(
                self._repeatFlag,
                self._description(optName),
                self._description(optName),
                " ".join(self._items),
            )
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteUsernames(Completer):
    """
    Complete usernames
    """

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            return f"{self._repeatFlag}:{self._description(optName)}:_users"
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteGroups(Completer):
    """
    Complete system group names
    """

    _descr = "group"

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            return f"{self._repeatFlag}:{self._description(optName)}:_groups"
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteHostnames(Completer):
    """
    Complete hostnames
    """

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            return f"{self._repeatFlag}:{self._description(optName)}:_hosts"
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteUserAtHost(Completer):
    """
    A completion action which produces matches in any of these forms::
        <username>
        <hostname>
        <username>@<hostname>
    """

    _descr = "host | user@host"

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            # Yes this looks insane but it does work. For bonus points
            # add code to grep 'Hostname' lines from ~/.ssh/config
            return (
                '%s:%s:{_ssh;if compset -P "*@"; '
                'then _wanted hosts expl "remote host name" _ssh_hosts '
                '&& ret=0 elif compset -S "@*"; then _wanted users '
                'expl "login name" _ssh_users -S "" && ret=0 '
                "else if (( $+opt_args[-l] )); then tmp=() "
                'else tmp=( "users:login name:_ssh_users -qS@" ) fi; '
                '_alternative "hosts:remote host name:_ssh_hosts" "$tmp[@]"'
                " && ret=0 fi}" % (self._repeatFlag, self._description(optName))
            )
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class CompleteNetInterfaces(Completer):
    """
    Complete network interface names
    """

    def _shellCode(self, optName, shellType):
        if shellType == _ZSH:
            return "{}:{}:_net_interfaces".format(
                self._repeatFlag,
                self._description(optName),
            )
        raise NotImplementedError(f"Unknown shellType {shellType!r}")


class Completions:
    """
    Extra metadata for the shell tab-completion system.

    @type descriptions: C{dict}
    @ivar descriptions: ex. C{{"foo" : "use this description for foo instead"}}
        A dict mapping long option names to alternate descriptions.  When this
        variable is defined, the descriptions contained here will override
        those descriptions provided in the optFlags and optParameters
        variables.

    @type multiUse: C{list}
    @ivar multiUse: ex. C{ ["foo", "bar"] }
        An iterable containing those long option names which may appear on the
        command line more than once. By default, options will only be completed
        one time.

    @type mutuallyExclusive: C{list} of C{tuple}
    @ivar mutuallyExclusive: ex. C{ [("foo", "bar"), ("bar", "baz")] }
        A sequence of sequences, with each sub-sequence containing those long
        option names that are mutually exclusive. That is, those options that
        cannot appear on the command line together.

    @type optActions: C{dict}
    @ivar optActions: A dict mapping long option names to shell "actions".
        These actions define what may be completed as the argument to the
        given option. By default, all files/dirs will be completed if no
        action is given. For example::

            {"foo"    : CompleteFiles("*.py", descr="python files"),
             "bar"    : CompleteList(["one", "two", "three"]),
             "colors" : CompleteMultiList(["red", "green", "blue"])}

        Callables may instead be given for the values in this dict. The
        callable should accept no arguments, and return a C{Completer}
        instance used as the action in the same way as the literal actions in
        the example above.

        As you can see in the example above. The "foo" option will have files
        that end in .py completed when the user presses Tab. The "bar"
        option will have either of the strings "one", "two", or "three"
        completed when the user presses Tab.

        "colors" will allow multiple arguments to be completed, separated by
        commas. The possible arguments are red, green, and blue. Examples::

            my_command --foo some-file.foo --colors=red,green
            my_command --colors=green
            my_command --colors=green,blue

        Descriptions for the actions may be given with the optional C{descr}
        keyword argument. This is separate from the description of the option
        itself.

        Normally Zsh does not show these descriptions unless you have
        "verbose" completion turned on. Turn on verbosity with this in your
        ~/.zshrc::

            zstyle ':completion:*' verbose yes
            zstyle ':completion:*:descriptions' format '%B%d%b'

    @type extraActions: C{list}
    @ivar extraActions: Extra arguments are those arguments typically
        appearing at the end of the command-line, which are not associated
        with any particular named option. That is, the arguments that are
        given to the parseArgs() method of your usage.Options subclass. For
        example::
            [CompleteFiles(descr="file to read from"),
             Completer(descr="book title")]

        In the example above, the 1st non-option argument will be described as
        "file to read from" and all file/dir names will be completed (*). The
        2nd non-option argument will be described as "book title", but no
        actual completion matches will be produced.

        See the various C{Completer} subclasses for other types of things which
        may be tab-completed (users, groups, network interfaces, etc).

        Also note the C{repeat=True} flag which may be passed to any of the
        C{Completer} classes. This is set to allow the C{Completer} instance
        to be re-used for subsequent command-line words. See the C{Completer}
        docstring for details.
    """

    def __init__(
        self,
        descriptions={},
        multiUse=[],
        mutuallyExclusive=[],
        optActions={},
        extraActions=[],
    ):
        self.descriptions = descriptions
        self.multiUse = multiUse
        self.mutuallyExclusive = mutuallyExclusive
        self.optActions = optActions
        self.extraActions = extraActions


def docMakeChunks(optList, width=80):
    """
    Makes doc chunks for option declarations.

    Takes a list of dictionaries, each of which may have one or more
    of the keys 'long', 'short', 'doc', 'default', 'optType'.

    Returns a list of strings.
    The strings may be multiple lines,
    all of them end with a newline.
    """

    # XXX: sanity check to make sure we have a sane combination of keys.

    # Sort the options so they always appear in the same order
    optList.sort(key=lambda o: o.get("short", None) or o.get("long", None))

    maxOptLen = 0
    for opt in optList:
        optLen = len(opt.get("long", ""))
        if optLen:
            if opt.get("optType", None) == "parameter":
                # these take up an extra character
                optLen = optLen + 1
            maxOptLen = max(optLen, maxOptLen)

    colWidth1 = maxOptLen + len("  -s, --  ")
    colWidth2 = width - colWidth1
    # XXX - impose some sane minimum limit.
    # Then if we don't have enough room for the option and the doc
    # to share one line, they can take turns on alternating lines.

    colFiller1 = " " * colWidth1

    optChunks = []
    seen = {}
    for opt in optList:
        if opt.get("short", None) in seen or opt.get("long", None) in seen:
            continue
        for x in opt.get("short", None), opt.get("long", None):
            if x is not None:
                seen[x] = 1

        optLines = []
        comma = " "
        if opt.get("short", None):
            short = "-%c" % (opt["short"],)
        else:
            short = ""

        if opt.get("long", None):
            long = opt["long"]
            if opt.get("optType", None) == "parameter":
                long = long + "="

            long = "%-*s" % (maxOptLen, long)
            if short:
                comma = ","
        else:
            long = " " * (maxOptLen + len("--"))

        if opt.get("optType", None) == "command":
            column1 = "    %s      " % long
        else:
            column1 = "  %2s%c --%s  " % (short, comma, long)

        if opt.get("doc", ""):
            doc = opt["doc"].strip()
        else:
            doc = ""

        if (opt.get("optType", None) == "parameter") and not (
            opt.get("default", None) is None
        ):
            doc = "{} [default: {}]".format(doc, opt["default"])

        if (opt.get("optType", None) == "parameter") and opt.get(
            "dispatch", None
        ) is not None:
            d = opt["dispatch"]
            if isinstance(d, CoerceParameter) and d.doc:
                doc = f"{doc}. {d.doc}"

        if doc:
            column2_l = textwrap.wrap(doc, colWidth2)
        else:
            column2_l = [""]

        optLines.append(f"{column1}{column2_l.pop(0)}\n")

        for line in column2_l:
            optLines.append(f"{colFiller1}{line}\n")

        optChunks.append("".join(optLines))

    return optChunks


def flagFunction(method, name=None):
    """
    Determine whether a function is an optional handler for a I{flag} or an
    I{option}.

    A I{flag} handler takes no additional arguments.  It is used to handle
    command-line arguments like I{--nodaemon}.

    An I{option} handler takes one argument.  It is used to handle command-line
    arguments like I{--path=/foo/bar}.

    @param method: The bound method object to inspect.

    @param name: The name of the option for which the function is a handle.
    @type name: L{str}

    @raise UsageError: If the method takes more than one argument.

    @return: If the method is a flag handler, return C{True}.  Otherwise return
        C{False}.
    """
    reqArgs = len(inspect.signature(method).parameters)
    if reqArgs > 1:
        raise UsageError("Invalid Option function for %s" % (name or method.__name__))
    if reqArgs == 1:
        return False
    return True


def portCoerce(value):
    """
    Coerce a string value to an int port number, and checks the validity.
    """
    value = int(value)
    if value < 0 or value > 65535:
        raise ValueError(f"Port number not in range: {value}")
    return value


portCoerce.coerceDoc = "Must be an int between 0 and 65535."  # type: ignore[attr-defined]



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Tests for L{twisted.web.websocket}.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Generic, TypeVar
from unittest import skipIf

from zope.interface import implementer

from twisted.internet.defer import Deferred
from twisted.internet.error import ConnectionDone
from twisted.internet.interfaces import IPushProducer
from twisted.internet.testing import AccumulatingProtocol, MemoryReactorClock
from twisted.python.failure import Failure
from twisted.test.iosim import ConnectionCompleter, IOPump
from twisted.trial.unittest import SynchronousTestCase
from twisted.web._responses import BAD_REQUEST
from twisted.web.client import Agent, readBody
from twisted.web.iweb import IRequest
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET, Request, Site
from twisted.web.static import Data

WSP = TypeVar("WSP", bound="WebSocketProtocol")
shouldSkip = False


class WeirdResource(Resource):
    def render_GET(self, request: Request) -> bytes:
        """
        Per U{the wsproto documentation
        <https://python-hyper.org/projects/wsproto/en/latest/api.html#wsproto.events.RejectConnection.has_body>}:

            - The only scenario in which the caller receives a RejectConnection
              with C{has_body == False} is if the peer violates sends an
              informational status code (1xx) other than 101

        This is a weird edge case so we provoke it.
        """
        request.setResponseCode(102)
        return b""


class DelayedResponse(Resource):
    """
    A resource that will not respond to a C{GET} request right away.  You will
    need to use the C{request.write(b'RESPONSE BODY')} and C{request.finish()}
    to trigger a response.
    """

    def render_GET(self, request: Request) -> int:
        self.request = request
        return NOT_DONE_YET


try:
    __import__("wsproto")
except ImportError:
    shouldSkip = True
else:
    from twisted.web.websocket import (
        ConnectionRejected,
        WebSocketClientEndpoint,
        WebSocketClientFactory,
        WebSocketProtocol,
        WebSocketResource,
        WebSocketServerFactory,
        WebSocketTransport,
    )

    @dataclass
    class MyWSP(WebSocketProtocol):
        """
        Used to implement both client-side and server-side WebSocket
        application that will help with testing.
        """

        pongs: list[bytes] = field(default_factory=list)
        wasLost: Failure | None = None

        def negotiationStarted(self, transport: WebSocketTransport) -> None:
            self.transport = transport

        def negotiationFinished(self) -> None:
            ...

        def connectionLost(self, reason: Failure) -> None:
            self.wasLost = reason

        def bytesMessageReceived(self, data: bytes | bytearray) -> None:
            if data == b"request":
                self.transport.sendBytesMessage(b"\x00resp\x01onse\xff")
            else:
                self.bDeferred.callback(data)

        def textMessageReceived(self, data: str) -> None:
            if data == "request":
                self.transport.sendTextMessage("response")
            else:
                self.deferred.callback(data)

        def pongReceived(self, payload: bytes) -> None:
            self.pongs.append(payload)

        def sendRequest(self) -> Deferred[str]:
            """
            Send a text message to the server and expect a response.
            """
            self.deferred: Deferred[str] = Deferred()
            self.transport.sendTextMessage("request")
            return self.deferred

        def bytesRequest(self) -> Deferred[bytes | bytearray]:
            """
            Send a bytes message to the server and expect a response.
            """
            self.bDeferred: Deferred[bytes | bytearray] = Deferred()
            self.transport.sendBytesMessage(b"request")
            return self.bDeferred

    class MyFactory(WebSocketServerFactory[MyWSP]):
        fixture: WebSocketFixture[Any]

        def buildProtocol(self, request: IRequest) -> MyWSP:
            new = MyWSP()
            self.fixture.servers.append(new)
            return new

    class MyClientFactory(WebSocketClientFactory[MyWSP]):
        def buildProtocol(self, uri: str) -> MyWSP:
            return MyWSP()


@dataclass
class WebSocketFixture(Generic[WSP]):
    clientFactory: WebSocketClientFactory[WSP] = field()
    reactor: MemoryReactorClock = field(default_factory=MemoryReactorClock)
    resource: Resource = field(default_factory=Resource)
    portNumber: int = 80
    servers: list[WSP] = field(default_factory=list)
    slowResource: DelayedResponse = field(default_factory=DelayedResponse)

    @classmethod
    def new(cls, clientFactory: WebSocketClientFactory[WSP]) -> WebSocketFixture[WSP]:
        self = cls(clientFactory)
        serverFactory = MyFactory()
        serverFactory.fixture = self
        self.resource.putChild(b"connect", WebSocketResource(serverFactory))
        self.resource.putChild(
            b"resource", Data(b"some-data", "application/octet-stream")
        )
        self.resource.putChild(b"processing", WeirdResource())
        self.resource.putChild(b"slow", self.slowResource)
        self.reactor.listenTCP(
            self.portNumber, Site(self.resource, reactor=self.reactor)
        )
        return self

    async def connect(self, uri: str = "http://localhost:80/connect") -> WSP:
        client = WebSocketClientEndpoint.new(self.reactor, uri)
        return await client.connect(self.clientFactory)

    def complete(self, greet: bool = True) -> IOPump:
        """
        There should be a single websocket connection in progress; complete it.

        @param greet: Should we immediately issue a greeting, i.e. deliver any
            pending buffered TCP data, upon connection?  If C{False}, the
            caller must call C{flush} on the result to deliver any pending
            data.
        """
        completer = ConnectionCompleter(self.reactor)
        succeeded = completer.succeedOnce(greet=greet)
        assert succeeded is not None, "Connection not in progress."
        return succeeded


@skipIf(shouldSkip, "wsproto library required for websockets")
class WebSocketTests(SynchronousTestCase):
    def test_websocket(self) -> None:
        """
        Connecting to a websocket server (installed with L{WebSocketResource})
        from a websocket client (connected with L{WebSocketClientEndpoint})
        results in a websocket connection.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        connected = Deferred.fromCoroutine(fixture.connect())
        self.assertNoResult(connected)
        self.assertEqual(len(fixture.reactor.tcpServers), 1)
        self.assertEqual(len(fixture.reactor.tcpClients), 1)
        pump = fixture.complete()
        wsClient = self.successResultOf(connected)
        requested = wsClient.sendRequest()
        self.assertNoResult(requested)
        pump.flush()
        self.assertEqual(self.successResultOf(requested), "response")

    def test_bytesMessage(self) -> None:
        """
        Connecting to a websocket server and sending it a bytes message results
        in C{bytesMessageReceived} being called.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        connected = Deferred.fromCoroutine(fixture.connect())
        pump = fixture.complete()
        wsClient = self.successResultOf(connected)
        bRequested = wsClient.bytesRequest()
        self.assertNoResult(bRequested)
        pump.flush()
        self.assertEqual(self.successResultOf(bRequested), b"\x00resp\x01onse\xff")

    def test_backpressure(self) -> None:
        """
        Websocket transports can notify a producer of backpressure events via
        attachProducer.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        Deferred.fromCoroutine(fixture.connect())
        pump = fixture.complete()
        events = []

        @implementer(IPushProducer)
        class TestProducer:
            def pauseProducing(self) -> None:
                events.append("paused")

            def resumeProducing(self) -> None:
                events.append("resumed")

            def stopProducing(self) -> None:
                events.append("stopped")

        testProducer = TestProducer()
        fixture.servers[0].transport.attachProducer(testProducer)
        self.assertEqual(events, [])
        pump.serverIO.producer.pauseProducing()
        self.assertEqual(events, ["paused"])
        pump.serverIO.producer.resumeProducing()
        self.assertEqual(events, ["paused", "resumed"])
        pump.serverIO.producer.stopProducing()
        self.assertEqual(events, ["paused", "resumed", "stopped"])
        self.assertIs(pump.serverIO.producer, testProducer)
        fixture.servers[0].transport.detachProducer()
        self.assertIs(pump.serverIO.producer, None)

    def test_pingPong(self) -> None:
        """
        Internally, Twisted's websocket resource responds to all ping requests
        with a pong as required by the spec, so peers can issue a C{ping} and
        receive a C{pong} message with the same payload; our websocket
        transport also implements a C{ping} method to send that message.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        connected = Deferred.fromCoroutine(fixture.connect())
        pump = fixture.complete()
        wsClient = self.successResultOf(connected)
        wsClient.transport.ping(b"123")
        self.assertEqual(wsClient.pongs, [])
        pump.flush()
        self.assertEqual(wsClient.pongs, [b"123"])

    def test_serverConnectionLost(self) -> None:
        """
        When the underlying TCP connection is lost,
        L{WebSocketProtocol.connectionLost} is invoked.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        connected = Deferred.fromCoroutine(fixture.connect())
        pump = fixture.complete()
        wsClient = self.successResultOf(connected)
        self.assertIs(fixture.servers[0].wasLost, None)
        self.assertIs(wsClient.wasLost, None)
        wsClient.transport.loseConnection()
        self.assertIs(fixture.servers[0].wasLost, None)
        self.assertIs(wsClient.wasLost, None)
        pump.flush()
        self.assertIsNot(fixture.servers[0].wasLost, None)
        self.assertIsNot(wsClient.wasLost, None)

    def test_bad(self) -> None:
        """
        Attempting to issue an C{HTTP GET} against a websocket server
        (installed with L{WebSocketResource}) results in a C{BAD_REQUEST}
        response.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        agent = Agent(fixture.reactor)
        response = agent.request(b"GET", b"http://localhost/connect")
        self.assertNoResult(response)
        fixture.complete()
        r = self.successResultOf(response)
        self.assertEqual(r.code, BAD_REQUEST)
        body = readBody(r)
        self.assertEqual(
            self.successResultOf(body), b"websocket protocol negotiation error"
        )

    def test_connectionRefused(self) -> None:
        """
        Attempting to connect to a regular HTTP resource that does not support
        websockets will result in the Deferred returned from
        L{WebSocketClientEndpoint} failing.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        connected = Deferred.fromCoroutine(fixture.connect("http://localhost/empty"))
        pump = fixture.complete(greet=False)
        self.assertNoResult(connected)
        pump.flush()
        self.failureResultOf(connected, ConnectionRejected)

    def test_connectionRefusedWeird(self) -> None:
        """
        Attempting to connect to a regular HTTP resource that does not support
        websockets will result in the Deferred returned from
        L{WebSocketClientEndpoint} failing.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        connected = Deferred.fromCoroutine(
            fixture.connect("http://localhost/processing")
        )
        pump = fixture.complete(greet=False)
        self.assertNoResult(connected)
        pump.flush()
        self.failureResultOf(connected, ConnectionRejected)

    def test_connectionRefusedSlow(self) -> None:
        """
        When the client connection is refused by an asynchronous HTTP response,
        the websocket client will not be notified until the response arrives.
        """
        fixture = WebSocketFixture.new(MyClientFactory())
        connected = Deferred.fromCoroutine(fixture.connect("http://localhost/slow"))
        pump = fixture.complete(greet=True)
        self.assertNoResult(connected)
        fixture.slowResource.request.write(b"")
        pump.flush()

        rejected = self.failureResultOf(connected, ConnectionRejected)
        resp = rejected.value.response
        resp.deliverBody(p := AccumulatingProtocol())
        self.assertEqual(p.data, b"")
        req = fixture.slowResource.request
        req.write(b"hello")
        pump.flush()
        self.assertEqual(p.data, b"hello")
        req.write(b"world")
        pump.flush()
        self.assertEqual(p.data, b"helloworld")
        req.finish()
        self.assertIs(p.closedReason, None)
        pump.flush()
        # expressed this way because assertIs isn't a guard for mypy
        assert p.closedReason is not None, "should be closed now"
        self.assertEqual(p.closedReason.type, ConnectionDone)



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Test L{twisted.web.pages}
"""

from typing import cast

from twisted.trial.unittest import SynchronousTestCase
from twisted.web.http_headers import Headers
from twisted.web.iweb import IRequest
from twisted.web.pages import errorPage, forbidden, notFound
from twisted.web.resource import IResource
from twisted.web.test.requesthelper import DummyRequest


def _render(resource: IResource) -> DummyRequest:
    """
    Render a response using the given resource.

    @param resource: The resource to use to handle the request.

    @returns: The request that the resource handled,
    """
    request = DummyRequest([b""])
    # The cast is necessary because DummyRequest isn't annotated
    # as an IRequest, and this can't be trivially done. See
    # https://github.com/twisted/twisted/issues/11719
    resource.render(cast(IRequest, request))
    return request


class ErrorPageTests(SynchronousTestCase):
    """
    Test L{twisted.web.pages._ErrorPage} and its public aliases L{errorPage},
    L{notFound} and L{forbidden}.
    """

    maxDiff = None

    def assertResponse(self, request: DummyRequest, code: int, body: bytes) -> None:
        self.assertEqual(request.responseCode, code)
        self.assertEqual(
            request.responseHeaders,
            Headers({b"content-type": [b"text/html; charset=utf-8"]}),
        )
        self.assertEqual(
            # Decode to str because unittest somehow still doesn't diff bytes
            # without truncating them in 2022.
            b"".join(request.written).decode("latin-1"),
            body.decode("latin-1"),
        )

    def test_escapesHTML(self) -> None:
        """
        The I{brief} and I{detail} parameters are HTML-escaped on render.
        """
        self.assertResponse(
            _render(errorPage(400, "A & B", "<script>alert('oops!')")),
            400,
            (
                b"<!DOCTYPE html>\n"
                b"<html><head><title>400 - A &amp; B</title></head>"
                b"<body><h1>A &amp; B</h1><p>&lt;script&gt;alert('oops!')"
                b"</p></body></html>"
            ),
        )

    def test_getChild(self) -> None:
        """
        The C{getChild} method of the resource returned by L{errorPage} returns
        the L{_ErrorPage} it is called on.
        """
        page = errorPage(404, "foo", "bar")
        self.assertIs(
            page.getChild(b"name", cast(IRequest, DummyRequest([b""]))),
            page,
        )

    def test_notFoundDefaults(self) -> None:
        """
        The default arguments to L{twisted.web.pages.notFound} produce
        a reasonable error page.
        """
        self.assertResponse(
            _render(notFound()),
            404,
            (
                b"<!DOCTYPE html>\n"
                b"<html><head><title>404 - No Such Resource</title></head>"
                b"<body><h1>No Such Resource</h1>"
                b"<p>Sorry. No luck finding that resource.</p>"
                b"</body></html>"
            ),
        )

    def test_forbiddenDefaults(self) -> None:
        """
        The default arguments to L{twisted.web.pages.forbidden} produce
        a reasonable error page.
        """
        self.assertResponse(
            _render(forbidden()),
            403,
            (
                b"<!DOCTYPE html>\n"
                b"<html><head><title>403 - Forbidden Resource</title></head>"
                b"<body><h1>Forbidden Resource</h1>"
                b"<p>Sorry, resource is forbidden.</p>"
                b"</body></html>"
            ),
        )



# -*- test-case-name: twisted.web.test.test_websocket -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Websocket (rfc6455) client and server support.

For websocket servers, place a L{WebSocketResource} into your Twisted Web
resource hierarchy.

For websocket clients, create a new endpoint via L{WebSocketClientEndpoint.new}
with the WebSocket server URL and then, on the newly created endpoint, call
L{WebSocketClientEndpoint.connect}.

Both client-side and server-side application code must conform to
L{WebSocketProtocol}.

@note: To use this module, you must install Twisted's C{websocket} extra, i.e.
    C{pip install twisted[websocket]}.
"""

from ._websocket_impl import (
    ConnectionRejected,
    WebSocketClientEndpoint,
    WebSocketClientFactory,
    WebSocketProtocol,
    WebSocketResource,
    WebSocketServerFactory,
    WebSocketTransport,
)

__all__ = [
    "ConnectionRejected",
    "WebSocketClientEndpoint",
    "WebSocketClientFactory",
    "WebSocketProtocol",
    "WebSocketResource",
    "WebSocketServerFactory",
    "WebSocketTransport",
]
